import asyncio
from bisect import bisect_left
import datetime as dt
import re
from typing import Dict, List
from optrabot.broker.optionpricedata import OptionStrikeData, OptionStrikePriceData
from optrabot.optionhelper import OptionHelper
from optrabot.broker.brokerconnector import BrokerConnector
from loguru import logger
from ib_async import *
import optrabot.config as optrabotcfg
from optrabot.models import Account
from optrabot.broker.order import Leg, OptionRight, Order as GenericOrder, OrderAction, OrderType, OrderStatus as GenericOrderStatus, Execution
from optrabot.tradetemplate.templatefactory import Template
import optrabot.symbolinfo as symbolInfo
from optrabot.managedtrade import ManagedTrade

class IBSymbolData:
	def __init__(self) -> None:
		self.symbol: str = None
		self.contract: Contract = None
		self.option_trading_class: str = None
		self.ticker: Ticker = None
		self.noPriceDataCount = 0
		self.lastPrice: float = None
		self.lastAtmStrike: float = None
		self.optionPriceData: Dict[dt.date, OptionStrikeData] = {}
		self.strikes: List[float] = []


class ForeignOrder(GenericOrder):
	"""
	Represents an order that has been placed by another user in the TWS.
	It is just used for managing conflicting orders.
	"""
	def __init__(self) -> None:
		super().__init__()

class IBTWSTConnector(BrokerConnector):
	def __init__(self) -> None:
		super().__init__()
		self._host = ''
		self._port = None
		self._clientid = None
		self._initialize()
		self.id = 'IBTWS'
		self.broker = 'IBKR'
		self._lock = asyncio.Lock()
		self._ib = IB()
		self._ib.errorEvent += self.onErrorEvent
		self._ib.disconnectedEvent += self.onDisconnected
		self._ib.execDetailsEvent += self.onExecDetailsEvent
		self._ib.orderStatusEvent += self.onOrderStatusEvent
		self._ib.commissionReportEvent += self.onCommissionReportEvent
		self._ib.pendingTickersEvent += self.onPendingTickers
		self._compiled_option_pattern = re.compile(pattern = r'^(?P<optionsymbol>[A-Z]+)\ +(?P<expiration>[0-9]+)(?P<type>[CP])(?P<strike>[0-9]+)')
		self._symbolData: Dict[str, IBSymbolData] = {}
		self._symbolReverseLookup: Dict[str, str] = {}		# maps IB symbol to generic symbol
		self._orders: List[GenericOrder] = []
		self._duplicate_order_ids: List[int] = []			# Liste mit Order-IDs, die als dopplete Order IDs von TWS gemeldet wurden
		self._conflictingOrderMap: Dict[GenericOrder,GenericOrder] = {}   # Maps the conflicting order to the causing order
		self._rule_details:  Dict[str, List[PriceIncrement]] = {}   # Map of Rule ID to Price Increment rules
		self._lock_rule = asyncio.Lock()

	def ___del__(self):
		"""
		Disconnect from TWS when object is deleted
		"""
		if self._ib.isConnected():
			self._ib.disconnect()

		# Unregister event handlers
		self._ib.errorEvent -= self.onErrorEvent
		self._ib.disconnectedEvent -= self.onDisconnected
		self._ib.execDetailsEvent -= self.onExecDetailsEvent
		self._ib.orderStatusEvent -= self.onOrderStatusEvent
		self._ib.commissionReportEvent -= self.onCommissionReportEvent
		self._ib.pendingTickersEvent -= self.onPendingTickers

	def _initialize(self):
		"""
		Initialize the TWS connector from the configuration
		"""
		config :optrabotcfg.Config = optrabotcfg.appConfig
		try:
			config.get('tws')
		except KeyError as keyErr:
			logger.debug('No TWS connection configured')
			return

		self._host = config.get('tws.host')
		if self._host == '':
			self._host = 'localhost'
		
		try:
			self._port = int(config.get('tws.port'))
		except KeyError as keyErr:
			self._port = 7496
		try:
			self._clientid = int(config.get('tws.clientid'))
		except KeyError as keyErr:
			self._clientid = 21
		self._initialized = True

	async def cancel_order(self, order: GenericOrder):
		""" 
		Cancels the given order
		"""
		await super().cancel_order(order)
		ibTrade: Trade = order.brokerSpecific['trade']
		if ibTrade.orderStatus.status != OrderStatus.Cancelled:
			self._ib.cancelOrder(ibTrade.order)

	async def connect(self):
		await super().connect()
		
		asyncio.create_task(self._connect_tws_task())

	async def disconnect(self):
		await super().disconnect()
		try:
			await self.eod_settlement_tasks()
			self._ib.disconnect()
		except Exception as excp:
			pass

	async def eod_settlement_tasks(self):
		"""
		Perform End of Day settlement tasks.
		- Unsubscribe from expired market data
		- Remove price data of expired options
		"""
		for symbolData in self._symbolData.values():
			try:
				today = dt.date.today()
				todays_option_price_data = symbolData.optionPriceData[today]
				for value in todays_option_price_data.strikeData.values():
					option_price_data: OptionStrikePriceData = value
					try:
						call_contract = option_price_data.brokerSpecific['call_contract']
						if call_contract:
							self._ib.cancelMktData(call_contract)
							option_price_data.brokerSpecific['call_contract'] = None
					except KeyError as keyErr:
						pass
					try:
						put_contract = option_price_data.brokerSpecific['put_contract']
						if put_contract:
							self._ib.cancelMktData(put_contract)
							option_price_data.brokerSpecific['put_contract'] = None
					except KeyError as keyErr:
						pass
				symbolData.optionPriceData.pop(today) # Delete the expired option price data
			except KeyError as keyErr:
				pass

	def getAccounts(self) -> List[Account]:
		"""
		Returns the accounts managed by the TWS connection
		"""
		if len(self._managedAccounts) == 0 and self.isConnected():
			for managedAccount in self._ib.managedAccounts():
				account = Account(id = managedAccount, name = managedAccount, broker = self.broker, pdt = False)
				self._managedAccounts.append(account)
		return self._managedAccounts

	def isConnected(self) -> bool:
		return self._ib.isConnected()
	
	async def prepareOrder(self, order: GenericOrder, need_valid_price_data: bool = True) -> bool:
		"""
		Prepares the given order for execution.
		- Retrieve current market data for order legs

		It returns True, if the order could be prepared successfully
		"""
		logger.debug("Wait for lock Prepare Order")
		async with self._lock:
			logger.debug("Got Prepare Order lock")
			try:
				symbolInformation = symbolInfo.symbol_infos[order.symbol]
				symbolData = self._symbolData[order.symbol]
				current_date = dt.date.today()
				#expiration = current_date.strftime('%Y%m%d')
				comboLegs: list[ComboLeg] = []
				comboContracts: list[Contract] = []
				option_price_data = symbolData.optionPriceData[current_date]
				for leg in order.legs:
					#legContract = Option(symbolData.contract.symbol, expiration, leg.strike, self._mappedRight(leg.right), symbolInformation.exchange, tradingClass = self._determineTradingClass(symbolData.contract.symbol))
					option_strike_price_data = option_price_data.strikeData[leg.strike]
					leg_contract = option_strike_price_data.brokerSpecific['call_contract'] if leg.right == OptionRight.CALL else option_strike_price_data.brokerSpecific['put_contract']
					#await self._ib.qualifyContractsAsync(legContract)
					if need_valid_price_data and not OptionHelper.checkContractIsQualified(leg_contract):
						logger.error("Contract {} is not qualified!", leg_contract)
						return False
					
					# Preisdaten für das Order Leg übernehmen
					if need_valid_price_data:
						if not option_strike_price_data.is_outdated():
							leg.askPrice = option_strike_price_data.putAsk if leg.right == OptionRight.PUT else option_strike_price_data.callAsk
							leg.bidPrice = option_strike_price_data.putBid if leg.right == OptionRight.PUT else option_strike_price_data.callBid
						else:
							logger.error("Price data for strike {} is outdated!", leg.strike)
							return False

					comboContracts.append(leg_contract)
					leg.brokerSpecific['contract'] = leg_contract
					comboLeg = ComboLeg(conId=leg_contract.conId, ratio=1, action=self._mappedOrderAction(leg.action), exchange=symbolInformation.exchange)
					comboLegs.append(comboLeg)
				order.brokerSpecific['comboLegs'] = comboLegs
			except Exception as excp:
				logger.error("Error preparing order: {}", excp)
				return False
			
			return True

	async def placeOrder(self, managed_trade: ManagedTrade, order: GenericOrder, parent_order: GenericOrder = None) -> bool:
		""" 
		Places the given order. It returns True if the entry order has been placed.
		"""
		logger.debug('Wait for Place order lock')
		async with self._lock:
			logger.debug('Got Place order lock')
			symbolInformation = symbolInfo.symbol_infos[order.symbol]
			symbolData = self._symbolData[order.symbol]		
	
			comboLegs = order.brokerSpecific['comboLegs']
			if not comboLegs:
				logger.error("Internal Error: Broker specific attribute comboContracts not filled!")
				return
			
			logger.debug('Checking for open Trades if Combo Legs are colliding')
			#contractsWithOpenOrders = await self._getAllOpenOrderContracts(template.account, symbolData.contract.symbol)

			conflictingOrders = await self._findConflictingOrders(order, managed_trade.template.account)
			if len(conflictingOrders) > 0:
				cancelledOCAs = []
				logger.debug(f'Found {len(conflictingOrders)} conflicting orders for symbol {order.symbol}...adding them to the conflicting order map')
				for conflictingIBTrade in conflictingOrders:
					conflictingOrder = self._findOrderByIBTrade(conflictingIBTrade)
					if conflictingOrder == None:
						logger.error(f'There is a foreign order from another user at the strike prices. Unable to handle it. Trade cannot be placed!')
						return

					self._conflictingOrderMap[conflictingOrder] = order # Conflicting Order mit der verursachenden Order verknüpfen und merken
					
					# Cancel the conflicting order
					if conflictingIBTrade.order.ocaGroup and conflictingIBTrade.order.ocaGroup in cancelledOCAs:
						# Order braucht nicht storniert werden, da bereits eine andere aus der gleichen OCA Gruppe storniert wurde
						pass
					else:
						self._ib.cancelOrder(conflictingIBTrade.order)
						if conflictingIBTrade.order.ocaGroup:
							cancelledOCAs.append(conflictingIBTrade.order.ocaGroup)

				# Urspüngliche Broker Order - Broker Specific [trade] ; Aktuelle Broker Order ; [Conflicting Trades]

				# Ich stormiere die Conflicting Trades
				# - Beim OrderStatusEvent darf die Stonierung für diese Orders nicht weitergreicht werden, weil sie in der Conflicting Trade Map
				#   enthalten sind

				# Es wird die aktuelle Order platziert
				# - Beim OrderStatusEvent "Filled" muss zusätzlich geprüft werden ob es in der Conflicting Trade Map einen Eintrag gibt
				#   wenn ja, dann müssen die referenzierten Conflicting Trades wieder in den Markt gelegt werden

			# Die Orders der behindernden Trades müssen hier storniert werden und deren Daten zwischengespeichert werden
			# Auch der Bezug zu dieser Order, denn wenn diese ausgeführt worden ist müssen die stornierten Orders wieder neu in den Markt gelegt werden.
			#for trade in conflictingTrades:
			#	trade.order.transmit = False
			#	trade.orderStatus.status = OrderStatus.Inactive
			#	self._ib.placeOrder(trade.contract, trade.order)

			#if len(conflictingTrades) > 0:
			#	logger.error(f'Order with opposite action already open for contract. Order not placed!')
			#	return False
			comboContract = None
			if len(order.legs) > 1:
				comboContract = Contract(symbol=symbolData.contract.symbol, secType='BAG', exchange=symbolInformation.exchange, currency=symbolData.contract.currency, comboLegs=comboLegs)
			else:
				comboContract = order.legs[0].brokerSpecific['contract']

			# For credit trades, the price must be negative
			if not managed_trade.long_legs_removed and (managed_trade.template.is_credit_trade() and order.price > 0) or (managed_trade.template.is_credit_trade() == False and order.price < 0):
				order.price = -order.price

			if order.type == OrderType.LIMIT:
				#order.price -= 0.50
				ibOrder = LimitOrder(self._mappedOrderAction(order.action), order.quantity, order.price)
				ibOrder.account = managed_trade.template.account
				ibOrder.outsideRth = True
			elif order.type == OrderType.STOP:
				ibOrder = StopOrder(self._mappedOrderAction(order.action), order.quantity, order.price)
				ibOrder.account = managed_trade.template.account
				ibOrder.outsideRth = True
			else:
				logger.error(f'Order type {order.type} currently not supported by IBKR Connector!')
			ibOrder.orderRef = order.orderReference
			if order.ocaGroup:  # https://interactivebrokers.github.io/tws-api/oca.html
				ibOrder.ocaGroup = order.ocaGroup
				ibOrder.ocaType = 1
			trade = self._ib.placeOrder(comboContract, ibOrder)
			order.brokerSpecific['comboContract'] = comboContract
			order.brokerSpecific['order'] = ibOrder
			order.brokerSpecific['trade'] = trade
			self._orders.append(order)
			logger.debug(f'Account: {managed_trade.template.account} Entry Order {ibOrder.orderId} placed for Trade: {trade} Number of contracts: {order.quantity}')
			return True

	async def place_complex_order(self, take_profit_order: GenericOrder, stop_loss_order: GenericOrder, template: Template) -> bool:
		"""
		TWS doesn't support complex orders
		"""
		raise NotImplementedError()
	
	async def adjustOrder(self, managed_trade: ManagedTrade, order: GenericOrder, price: float) -> bool:
		""" 
		Adjusts the given order with the given new price
		"""
		if order.status == GenericOrderStatus.FILLED:
			logger.info('Order {} is already filled. Adjustment not required.', order)
			return True
		if order.status == GenericOrderStatus.CANCELLED:
			logger.info('Order {} is already cancelled. Adjustment not possible.', order)
			return True

		# For credit trades, the price must be negative
		if not managed_trade.long_legs_removed and (managed_trade.template.is_credit_trade() and order.price > 0) or (managed_trade.template.is_credit_trade() == False and order.price < 0):
			price = -price		

		trade: Trade = order.brokerSpecific['trade']
		try:
			comboContract = order.brokerSpecific['comboContract']
			ibOrder: Order = order.brokerSpecific['order']
			if order.type == OrderType.LIMIT:
				ibOrder.lmtPrice = price
			elif order.type == OrderType.STOP:
				ibOrder.auxPrice = price
			trade = self._ib.placeOrder(comboContract, ibOrder)
			order.brokerSpecific['trade'] = trade
			return True
		except Exception as excp:
			logger.error('Exception beim Anpassen der Order')
		return False

	async def requestTickerData(self, symbols: List[str]):
		"""
		Request ticker data for the given symbols and their options
		"""
		contracts = []
		for symbol in symbols:
			requestOptionsData = False
			match symbol:
				case 'SPX':
					symbolData = IBSymbolData()
					symbolData.symbol = symbol
					symbolInformation = symbolInfo.symbol_infos[symbolData.symbol]
					symbolData.contract = Index('SPX', 'CBOE')
					symbolData.option_trading_class = symbolInformation.trading_class
					self._symbolData[symbol] = symbolData
					self._symbolReverseLookup[symbolData.contract.symbol] = symbol
					contracts.append(symbolData.contract)
				case 'VIX':
					symbolData = IBSymbolData()
					symbolData.symbol = symbol
					symbolData.contract = Index('VIX', 'CBOE')
					self._symbolData[symbol] = symbolData
					self._symbolReverseLookup[symbolData.contract.symbol] = symbol
					contracts.append(symbolData.contract)
				case _:
					logger.error(f'Symbol {symbol} currently not supported by IBKR Connector!')
					continue

		try:
			qualifiedContracts = await self._ib.qualifyContractsAsync(*contracts)
			for contract in qualifiedContracts:
				if contract.symbol != 'VIX':  # Options not relevant for VIX
					symbolInformation = symbolInfo.symbol_infos[contract.symbol]
					chains = await self._ib.reqSecDefOptParamsAsync(contract.symbol, '', contract.secType, contract.conId)
					if len(chains) > 0:
						chain = next(c for c in chains if c.tradingClass == symbolInformation.trading_class and c.exchange == contract.exchange)
						if chain == None:
							logger.error(f'No Option Chain for trading class {symbolInformation.trading_class} found. Unable to request option data')

						current_date = dt.date.today()
						expiration = current_date.strftime('%Y%m%d')
						symbolData = self._symbolData[contract.symbol]
						symbolData.strikes = chain.strikes
						if int(chain.expirations[0]) > int(expiration):
							logger.warning(f'There are no {contract.symbol} options expiring today!')
				
				self._ib.reqMktData(contract, '', False, False)
		except Exception as excp:
			logger.error(f'Error qualifying contract {symbolData.contract}: {excp}')
			return
			
			#self._ib.reqMktData(symbolData.contract, '', False, False)

			# if requestOptionsData:
			# 	chains = await self._ib.reqSecDefOptParamsAsync(symbolData.contract.symbol, '', symbolData.contract.secType, symbolData.contract.conId)
			# 	chain = next(c for c in chains if c.tradingClass == optiontradingclass and c.exchange == 'CBOE')
			# 	if chain == None:
			# 		logger.error(f'No Option Chain for trading class {optiontradingclass} found. Unable to request option data')
			# 		return
				
			# 	current_date = dt.date.today()
			# 	expiration = current_date.strftime('%Y%m%d')

			# 	if int(chain.expirations[0]) > int(expiration):
			# 		logger.warning('There are no SPX options expiring today!')
			# 		return

	async def _connect_tws_task(self):
		try:
			await self._ib.connectAsync(self._host, self._port, clientId=self._clientid)
			self._emitConnectedEvent()
			
		except Exception as excp:
			self._emitConnectFailedEvent()
			#logger.error("Error connecting TWS: {}", excp)
			#attempt += 1
			#logger.error('Connect failed. Retrying {}. attempt in {} seconds', attempt, delaySecs)
			#await asyncio.sleep(delaySecs)
			#asyncio.create_task(self._connect_ib_task(attempt, delaySecs))

	async def onDisconnected(self):
		#logger.warning('Disconnected from TWS, attempting to reconnect in 30 seconds ...')
		self._tradingEnabled = False
		#asyncio.create_task(self._reconnect_ib_task())
		self._emitDisconnectedEvent()

	async def onErrorEvent(self, reqId: int, errorCode: int, errorString: str, contract: Contract):
		if errorCode in {104, 200, 202, 399, 2103, 2104, 2105, 2106, 2107, 2108, 2109, 2157, 2158}:
			# 104: Cannot modify a filled order
			# 200: No security definition has been found for the request, contract
			# 202: Order wurde storniert z.B. TP order bei OCO, wenn SL Order ausgeführt wurde
			# 399: Warnung das die Order nur während der regulären Handelszeiten ausgeführt wird
			# 10147: OrderId ... that needs to be cancelled is not found
			# 2103, 2104, 2105, 2106, 2108, 2158: Marktdatenverbindung ignorieren
			# 2107: Die Verbindung zum HMDS-Datenzentrum is inaktv, sollte jedoch auf Anfrage verfügbar seind
			# 2109: Warnhinweis zu einem Orderereignis außerhalb der regulären Handelszeiten
			# 2157: Verbindung zum Sec-def-Datenzentrum unterbrochen
			return
		elif errorCode == 201:
			# 201: Order rejected - reason:Stop price revision is disallowed after order has triggered
			# 201: Order rejected - reason:We are unable to accept your order. Your Available Funds are in sufficient
			# 201: Order abgewiesen - Grund:Order has been cancelled already, too late to replace
			if errorString.find('Margin') > -1:
				logger.error('Order rejected: Insufficient margin!')
				return

		elif errorCode == 10147:
			# Die zu stornierende Order wurde nicht gefunden
			if reqId in self._duplicate_order_ids:
				logger.info(f'Order {reqId}' + ' is a duplicate order ID. Ignoring error event.')
				return

		elif errorCode == 10148:
			# Order that should be cancelled cannot be cancelled
			# Passiert beim Cancel einer TP oder SL Order, wenn sie per OCO Group verbunden sind und eine der Orders bereits
			# storniert wurde

			# Prüfen ob die betreffende Order in den conflicting Orders enthalten ist
			for entry in self._conflictingOrderMap.keys():
				order: GenericOrder = entry
				if order.brokerSpecific['trade'].order.orderId == reqId:
					logger.debug(f'Order {reqId} is conflicting with another order. Ignoring error event.')
					return
		
		elif errorCode == 10168:
			# Für die angeforderten Markdaten besteht kein Abonnement
			logger.error(f'No valid market data subscription for symbol {contract.symbol}')
			self._tradingEnabled = False
			return

		elif errorCode == 1100:
			# Connection between TWS and IB lost.
			logger.warning('Connection between TWS and Interactive Brokers lost -> Trading disabled!')
			self._tradingEnabled = False
			return
		elif errorCode == 1102:
			# Connection between TWS and IB restored
			logger.success('Connection between TWS and Interactive Brokers has been reestablished! -> Trading enabled!')
			self._tradingEnabled = True
			return
		elif errorCode == 2148:
			# Margin Warning
			logger.warning('IB reports potential margin violation!')
			return

		errorData = {'action': 'errorEvent','reqId':reqId, 'errorCode':errorCode, 'errorString':errorString, 'contract': contract}
		logger.error('IB raised following error: {}', errorData)
	
	async def onExecDetailsEvent(self, trade: Trade, fill: Fill):
		""" This eventhandler is called on trade execution.
		The event is raised for each single fill of an order leg
		"""
		logger.debug('Exec Detail for trade {}', trade)
		logger.debug('Fill: {}', fill)

		for order in self._orders:
			brokerSpecificTrade: Trade = order.brokerSpecific['trade']
			if brokerSpecificTrade.order.orderId == trade.order.orderId:
				relevantOrder = order
				break
		if relevantOrder == None:
			logger.debug(f'No managed order matched the execution details event')
			return
		if fill.contract.secType == 'BAG':
			# Do not store inform stion on combo executions
			return
		if fill.execution.side == 'BOT':
			action = OrderAction.BUY
		else:
			action = OrderAction.SELL
		expirationDate = dt.datetime.strptime(fill.contract.lastTradeDateOrContractMonth, '%Y%m%d')
		execution = Execution(id=fill.execution.execId, 
							action=action,
							sec_type=fill.contract.right,
							strike=fill.contract.strike,
							amount=fill.execution.shares,
							price=fill.execution.avgPrice,
							expiration=expirationDate,
							timestamp=fill.execution.time)
		self._emitOrderExecutionDetailsEvent(relevantOrder, execution)

	async def onCommissionReportEvent(self, trade: Trade, fill: Fill, report: CommissionReport):
		"""
		Handles the Commission Report Event
		"""
		relevantOrder = None
		logger.debug('Commission Report order id {} and fill {}', trade.order.orderId, fill)
		logger.debug('Commission Report: {}', report)
		for order in self._orders:
			brokerSpecificTrade: Trade = order.brokerSpecific['trade']
			if brokerSpecificTrade.order.orderId == trade.order.orderId:
				relevantOrder = order
				break
		if relevantOrder == None:
			logger.debug(f'No managed order matched the commission report event')
			return

		self._emitCommissionReportEvent(relevantOrder, fill.execution.execId, report.commission, 0)

	async def onOrderStatusEvent(self, trade: Trade):
		"""
		Handles the Order Status Event
		"""
		logger.debug(f'onOrderStatusEvent() Order: {trade.order.orderId} Status: {trade.orderStatus.status}')
		relevantOrder: GenericOrder = None
		for order in self._orders:
			brokerSpecificTrade: Trade = order.brokerSpecific['trade']
			if brokerSpecificTrade.order.orderId == trade.order.orderId:
				relevantOrder = order
				break
		if relevantOrder == None:
			logger.debug(f'No managed order matched the status event')
			return
		
		# Prüfen ob die order in der Conflicting Order Map enthalten ist
		if relevantOrder in self._conflictingOrderMap.keys():
			causingOrder = self._conflictingOrderMap[relevantOrder]
			causingOrderOrderId = causingOrder.brokerSpecific['trade'].order.orderId
			logger.debug(f'Order {trade.order.orderId} is conflicting with order {causingOrderOrderId}. Ignoring status event.')
			return

		if trade.orderStatus.status == OrderStatus.Cancelled or trade.orderStatus.status == OrderStatus.Inactive or trade.orderStatus.status == OrderStatus.ApiCancelled:
			log_entry: TradeLogEntry = None
			for log_entry in trade.log:
				if log_entry.status != 'Cancelled':
					continue
				logger.debug(f'Order Log: {log_entry}')
			if log_entry == None:
				logger.error(f'No log entry found for cancelled order!')
			elif log_entry.errorCode == 103:
				# Error 103, reqId 10292: Doppelt vorhandene Order-ID
				logger.warning('Adjustment of entry order has been rejected, because Duplicate Order-ID.')
				duplicate_order_id = trade.order.orderId
				self._duplicate_order_ids.append(duplicate_order_id)
				logger.debug(f'Added Order ID {duplicate_order_id} to duplicate order IDs')
				# In diesem Fall muss die Order sicherheitshalber neu platziert werden
				try:
					self._ib.cancelOrder(trade.order)
				except Exception as excp:
					logger.debug(f'Error cancelling order: {excp}')
				trade.order.orderId = self._ib.client.getReqId()
				logger.debug(f'Replacing order with new Order Id {trade.order.orderId}')
				relevantOrder.brokerSpecific['trade'] = self._ib.placeOrder(trade.contract, trade.order)
				logger.info(f'Order {duplicate_order_id} has been replaced with new Order Id {trade.order.orderId}')
				return
			elif log_entry.errorCode == 104:
				# Error 104: Cannot modify a filled order
				logger.warning(f'Order {trade.order.orderId} has been filled already. Ignoring the Cancelled status event.')
				return
			
			# Ingnore the Cancelled status event for duplicate order IDs
			if trade.order.orderId in self._duplicate_order_ids:
				logger.debug(f'Order {trade.order.orderId} is a duplicate order ID. Ignoring the Cancelled status event.')
				return
			else:
				logger.debug(f'Order with ID {trade.order.orderId} is not in the list of duplicate order IDs')

			# Wenn die stornierte Order eine verursachende Order für Conflicting Ordrer(s) ist, dann müssen die stornierten Orders wieder in den Markt gelegt werden
			await self._placeConflictingOrdersAgain(relevantOrder)
		
		filledAmount = 0
		if trade.orderStatus.status == OrderStatus.Filled:
			# Wenn die gefüllte Order eine verursachende Order für conflicting Order(s) ist und sie vollständig ausgeüfhrt wurde,
			# dann müssen die stornierten Orders wieder in den Markt gelegt werden
			await self._placeConflictingOrdersAgain(relevantOrder)

			filledAmount = relevantOrder.quantity - trade.remaining()
			relevantOrder.averageFillPrice = trade.orderStatus.avgFillPrice

		self._emitOrderStatusEvent(relevantOrder, self._genericOrderStatus(trade.orderStatus.status), filledAmount)
		
	def getFillPrice(self, order: GenericOrder) -> float:
		""" 
		Returns the fill price of the given order if it is filled
		"""
		trade: Trade = order.brokerSpecific['trade']
		return abs(trade.orderStatus.avgFillPrice)
	
	def getLastPrice(self, symbol: str) -> float:
		""" 
		Returns the last price of the given symbol
		"""
		symbolData = self._symbolData[symbol]
		return symbolData.lastPrice

	def get_option_strike_price_data(self, symbol: str, expiration: dt.date, strike: float) -> OptionStrikePriceData:
		""" 
		Returns the option strike price data for the given symbol, expiration date, strike price and right
		"""
		symbolData = self._symbolData[symbol]
		try:
			optionStrikeData = symbolData.optionPriceData[expiration]
			if strike in optionStrikeData.strikeData.keys():
				return optionStrikeData.strikeData[strike]
			else:
				return None
		except KeyError as keyErr:
			return None

	def get_strike_by_delta(self, symbol: str, right: str, delta: int) -> float:
		""" 
		Returns the strike price based on the given delta based on the buffered option price data
		"""
		symbolData = self._symbolData[symbol]
		current_date = dt.date.today()
		option_price_data: OptionStrikeData= symbolData.optionPriceData[current_date]
		previous_delta = 0
		previous_strike = 0
		reverse = True if right == OptionRight.PUT else False
		sorted_strikes = dict(sorted(option_price_data.strikeData.items(), reverse=reverse))
		for strike, price_data in sorted_strikes.items():
			if right == OptionRight.PUT:
				if price_data.putDelta == None:
					continue
				adjusted_delta = price_data.putDelta * -100
			else:
				if price_data.callDelta == None:
					continue
				adjusted_delta = price_data.callDelta * 100
			if adjusted_delta <= delta:
				if OptionHelper.closest_number(delta, previous_delta, adjusted_delta) == adjusted_delta:
					return strike
				else:
					return previous_strike
			previous_delta = adjusted_delta
			previous_strike = strike

		raise ValueError(f'No strike price found for delta {delta} in symbol {symbol}!')

	async def onPendingTickers(self, tickers: List[Ticker]):
		"""
		Handles the pending tickers event
		"""
		for ticker in tickers:
			if ticker.contract.symbol in self._symbolData.keys():
				logger.trace(f'Ticker for symbol {ticker.contract.symbol} received.')
				self._tradingEnabled = True
				ticker.lastExchange
				logger.trace(f'Ticker Last {ticker.last}: {ticker}')
				symbolData = self._symbolData[ticker.contract.symbol]
				symbolData.ticker = ticker
				if ticker.last and util.isNan(ticker.last) == False:
					symbolData.lastPrice = ticker.last
				else:
					symbolData.lastPrice = ticker.close

				# Prüfen ob Preisdaten vorhanden sind
				if util.isNan(symbolData.lastPrice) and ticker.contract.secType != 'OPT':
					# Wenn mehrmals keine gültigen Preisdaten empfangen wurden, dann wird der Handel deaktiviert
					symbolData.noPriceDataCount += 1
					if symbolData.noPriceDataCount > 5:
						logger.error(f'Receiving no valid price data for symbol {symbolData.symbol}. Trading disabled!')
						self._tradingEnabled = False
						break
					continue
				else:
					symbolData.noPriceDataCount = 0
					self._tradingEnabled = True

				# Für VIX werden keine weiteren Daten gelesen
				if ticker.contract.symbol == 'VIX':
					break

				if ticker.contract.secType != 'OPT':
					# Keine Optionen -> Fehlende Optionen abfragen
					try:
						atmStrike = OptionHelper.roundToStrikePrice(symbolData.lastPrice)
						if symbolData.lastAtmStrike != atmStrike:  # Check for missing Option Data only if ATM Strike has changed
							expirationDate = dt.date.today()
							symbolData.lastAtmStrike = atmStrike
							await self._requestMissingOptionData(symbolData, expirationDate, atmStrike)
					except KeyError as keyErr:
						logger.error(f'No generic symbol found for IB TWS symbol {ticker.contract.symbol}')
				else:
					# Option verarbeiten
					try:
						optionType, expiration, strike = self._getOptionInfos(ticker.contract)
					except Exception as exc:
						logger.error(f'Error getting option infos: {exc}')
					genericSymbol = self._symbolReverseLookup[ticker.contract.symbol]
					optionStrikePriceData = self.get_option_strike_price_data(genericSymbol, expiration, strike)
					if optionStrikePriceData:
						if optionType == OptionRight.CALL:
							optionStrikePriceData.callAsk = ticker.ask
							optionStrikePriceData.callBid = ticker.bid
							if ticker.modelGreeks:
								optionStrikePriceData.callDelta = ticker.modelGreeks.delta
						else:
							optionStrikePriceData.putAsk = ticker.ask
							optionStrikePriceData.putBid = ticker.bid
							if ticker.modelGreeks:
								optionStrikePriceData.putDelta = ticker.modelGreeks.delta
						optionStrikePriceData.lastUpdated = dt.datetime.now()
					else:
						logger.debug(f'No option data available for strike {strike} in symbol {genericSymbol} yet. Ignoring Ticker data.')
					
	def uses_oco_orders(self) -> bool:
		""" 
		The TWS Connector uses OCO orders for take profit and stop loss orders
		"""
		return True
	
	def _createReplacementOrder(self, ibOrder: Order) -> Order:
		"""
		Creates a new replacement order for the given order
		"""
		if isinstance(ibOrder, LimitOrder):
			replacementOrder = LimitOrder(ibOrder.action, ibOrder.totalQuantity, ibOrder.lmtPrice)
		elif isinstance(ibOrder, StopOrder):
			replacementOrder = StopOrder(ibOrder.action, ibOrder.totalQuantity, ibOrder.auxPrice)
		else:
			logger.error(f'Order type {type(ibOrder)} currently not supported by IBKR Connector!')
			return None
		replacementOrder.account = ibOrder.account
		replacementOrder.outsideRth = ibOrder.outsideRth
		replacementOrder.orderRef = ibOrder.orderRef
		replacementOrder.ocaGroup = ibOrder.ocaGroup
		replacementOrder.ocaType = ibOrder.ocaType
		replacementOrder.transmit = True
		return replacementOrder

	def _findOrderByIBTrade(self, trade: Trade) -> GenericOrder:
		"""
		Find the generic order that belongs to the given IBKR trade
		"""
		for order in self._orders:
			if order.brokerSpecific['trade'] == trade:
				return order
		return None

	async def _findConflictingOrders(self, order: GenericOrder, account: str) -> List[Trade]:
		"""
		Find all open trades for the given account and symbol that are conflicting with the given order
		"""
		comboLegs = order.brokerSpecific['comboLegs']
		oppositeAction = 'BUY' if self._mappedOrderAction(order.action) == 'SELL' else 'SELL'
		conflictingTrades: List[Trade] = []
		openTrades: List[Trade] = await self._ib.reqAllOpenOrdersAsync()
		
		for openTrade in openTrades:
			if openTrade.order.transmit == False or openTrade.contract.symbol != order.symbol or openTrade.order.account != account or openTrade.order.action != oppositeAction:
				continue

			# ausgeführte und stonierte Trades/Orders spielen keine Rolle (sollten eigentlich nicht mehr in der Liste sein)
			if openTrade.orderStatus.status == OrderStatus.Cancelled or openTrade.orderStatus.status == OrderStatus.Filled:
				continue

			# Offene Order mit entgegengesetzter Aktion gefunden -> Sind enhalten sie Kontrakte die mit der neuen Order kollidieren
			if openTrade.contract.secType == 'BAG':
				for leg in openTrade.contract.comboLegs:
					for comboLeg in comboLegs:
						if leg.conId == comboLeg.conId:
							if openTrade not in conflictingTrades:
								conflictingTrades.append(openTrade)
			else:
				for comboLeg in comboLegs:
					if openTrade.contract.conId == comboLeg.conId:
						if openTrade not in conflictingTrades:
							conflictingTrades.append(openTrade)
		return conflictingTrades

	def _genericOrderStatus(self, status: OrderStatus) -> GenericOrderStatus:
		"""
		Maps the IBKR specific order status to the general order status
		"""
		match status:
			case OrderStatus.PendingSubmit:
				return GenericOrderStatus.OPEN
			case OrderStatus.PreSubmitted:
				return GenericOrderStatus.OPEN
			case OrderStatus.Submitted:
				return GenericOrderStatus.OPEN
			case OrderStatus.ApiCancelled:
				return GenericOrderStatus.CANCELLED
			case OrderStatus.Cancelled:
				return GenericOrderStatus.CANCELLED
			case OrderStatus.Filled:
				return GenericOrderStatus.FILLED
			case OrderStatus.Inactive:
				return GenericOrderStatus.CANCELLED
			case OrderStatus.PendingCancel:
				return GenericOrderStatus.OPEN
			case OrderStatus.Cancelled:
				return GenericOrderStatus.CANCELLED

	def _getOptionInfos(self, option: Option) -> tuple:
		"""
		Extracts the generic symbol and expiration date, strike price from the IBKR option contract
		"""
		error = False
		match = self._compiled_option_pattern.match(option.localSymbol)
		try:
			if match:
				for symbol, symbol_info in symbolInfo.symbol_infos.items():
					if symbol_info.symbol == option.symbol:
						genericSymbol = symbol_info.symbol
						break
				expiration = dt.datetime.strptime(match.group('expiration'), '%y%m%d').date()
				optionType = OptionRight.CALL if match.group('type') == 'C' else OptionRight.PUT
			else:
				raise IndexError
		except IndexError as indexErr:
			logger.error(f'Invalid option symbol {option.localSymbol}')
			error = True
		except ValueError as valueErr:
			logger.error(f'Invalid option symbol {option.localSymbol}')
			error = True
		if genericSymbol == None or error == True:
			raise ValueError(f'Invalid option symbol {option.localSymbol}')
		return optionType, expiration, option.strike

	def _mappedOrderAction(self, action: OrderAction) -> str:
		"""
		Maps the general order action to the IBKR specific order action
		"""
		match action:
			case OrderAction.BUY:
				return 'BUY'
			case OrderAction.BUY_TO_OPEN:
				return 'BUY'
			case OrderAction.BUY_TO_CLOSE:
				return 'BUY'
			case OrderAction.SELL:
				return 'SELL'
			case OrderAction.SELL_TO_OPEN:
				return 'SELL'
			case OrderAction.SELL_TO_CLOSE:
				return 'SELL'
			case _:
				logger.error(f'Order action {action} not supported by IBKR Connector!')
				return None
	
	def _mappedRight(self, right: OptionRight) -> str:
		"""
		Maps the general option right to the IBKR specific option right
		"""
		match right:
			case OptionRight.CALL:
				return 'C'
			case OptionRight.PUT:
				return 'P'
			case _:
				raise ValueError(f'Option right {right} not supported by IBKR Connector!')
				return None
			
	async def _placeConflictingOrdersAgain(self, order: GenericOrder):
		"""
		Relevante temporär stornierte Conflicting Orders wieder in den Markt legen, wenn
		die übergebene Order eine verursachende Order ist
		"""
		resolvedConflictingOrders: List[GenericOrder] = []
		for key, value in self._conflictingOrderMap.items():
			if value == order:
				logger.debug(f'Causing order of a conflicting order has been filled. Re-placing conflicting order again.')
				if type(key) == ForeignOrder:
					logger.debug(f'Conflicting order is a foreign order. Re-placing it.')
				conflictingOrder: GenericOrder = key
				ibOrder:Order = conflictingOrder.brokerSpecific['order']
				replacementOrder = self._createReplacementOrder(ibOrder)
				replacementTrade = self._ib.placeOrder(conflictingOrder.brokerSpecific['comboContract'], replacementOrder)
				logger.debug(f'Conflicting order {ibOrder.orderId} has been re-placed with order id {replacementTrade.order.orderId}')
				conflictingOrder.brokerSpecific['order'] = replacementOrder
				conflictingOrder.brokerSpecific['trade'] = replacementTrade
				resolvedConflictingOrders.append(conflictingOrder)
		
		# Jetzt wiedereingestellte conflicting orders aus der Map entfernen
		for entry in resolvedConflictingOrders:
			self._conflictingOrderMap.pop(entry)

	async def _getAllOpenOrderContracts(self, account, symbol) -> List:
		"""
		Determine the ContractIds of all open Orders for the given account and symbol
		"""
		openTrades: List[Trade] = await self._ib.reqAllOpenOrdersAsync()
		openOrderContracts = list()
		for openTrade in openTrades:
			if openTrade.contract.symbol != symbol or openTrade.order.account != account:
				continue
			if openTrade.contract.secType == 'BAG':
				for leg in openTrade.contract.comboLegs:
					openOrderContracts.append(leg.conId)
			else:
				openOrderContracts.append(openTrade.contract.conId)
		return openOrderContracts
	
	def _calculateMidPrice(self, legs: List[Leg], tickers: List[Ticker], contracts: List[Contract]) -> float:
		"""
		Calculates the mid price for the given tickers
		"""
		midPrice = None
		for leg in legs:
			for ticker in tickers:
				if ticker.contract.strike == leg.strike and ticker.contract.right == self._mappedRight(leg.right):
					leg
					legMidPrice = (ticker.ask + ticker.bid) / 2
					if util.isNan(legMidPrice) or (ticker.ask == -1.00 and ticker.bid == -1.00):
						if leg.action == OrderAction.BUY:
							legMidPrice = 0.05
						else:
							return None
					if leg.action == OrderAction.SELL:
						midPrice = -legMidPrice
					else:
						midPrice += legMidPrice
					break

		return OptionHelper.roundToTickSize(midPrice)
	
	async def _requestMissingOptionData(self, symbolData: IBSymbolData, expirationDate: dt.date, atmStrike: float):
		"""
		Request option data for the given symbol and expiration date
		"""
		# Bestehende gespeicherte Optionen abfragen
		try:
			expiration_date_str = expirationDate.strftime('%Y%m%d')
			optionStrikeData = symbolData.optionPriceData[expirationDate]
		except KeyError as keyErr:
			# Wenn noch keine Optionsdaten für das Verfallsdatum vorhanden sind
			# dann bei der TWS anfragen ob es Optionen gibt
			chains = await self._ib.reqSecDefOptParamsAsync(symbolData.contract.symbol, '', symbolData.contract.secType, symbolData.contract.conId)
			try:
				chain = next(c for c in chains if c.tradingClass == symbolData.option_trading_class and c.exchange == symbolData.contract.exchange)
				if int(chain.expirations[0]) > int(expiration_date_str):
					return
				
				optionStrikeData = OptionStrikeData()
				symbolData.optionPriceData[expirationDate] = optionStrikeData
			
			except Exception as excp:
				logger.error(f'Unexpected error: {excp}')
				return
			pass

		# Die 25 Strikes um den ATM Strike herum abfragen
		symbolInformation = symbolInfo.symbol_infos[symbolData.symbol]
		number_of_strikes_above = 10
		number_of_strikes_below = 25
		pos = bisect_left(symbolData.strikes, atmStrike)
		lower_bound = max(0, pos - number_of_strikes_below)
		upper_bound = min(len(symbolData.strikes), pos + number_of_strikes_above)
		strikes_of_interest = symbolData.strikes[lower_bound:upper_bound]

		options_to_be_requested = []
		for strikePrice in strikes_of_interest:
			try:
				optionStrikeData.strikeData[strikePrice]
			except KeyError as keyErr:
				option_strike_price_data = OptionStrikePriceData()
				call_option_to_be_requested = Option(symbolData.symbol , expiration_date_str, strikePrice, self._mappedRight(OptionRight.CALL), symbolInformation.exchange , tradingClass=symbolData.option_trading_class)
				list_of_details = await self._ib.reqContractDetailsAsync(call_option_to_be_requested)
				if len(list_of_details) == 0:
					logger.info(f'No Options available at strike {strikePrice}')
					continue

				optionStrikeData.strikeData[strikePrice] = option_strike_price_data

				put_option_to_be_requested = Option(symbolData.symbol , expiration_date_str, strikePrice, self._mappedRight(OptionRight.PUT), symbolInformation.exchange , tradingClass=symbolData.option_trading_class)

				option_strike_price_data.brokerSpecific['call_contract'] = call_option_to_be_requested
				options_to_be_requested.append(call_option_to_be_requested)
				
				option_strike_price_data.brokerSpecific['put_contract'] = put_option_to_be_requested
				options_to_be_requested.append(put_option_to_be_requested)
		
		if len(options_to_be_requested) > 0:
			qualifiedContracts = await self._ib.qualifyContractsAsync(*options_to_be_requested)
			for qualified_contract in qualifiedContracts:
				self._ib.reqMktData(qualified_contract, '', False, False)


from abc import ABC, abstractmethod
from datetime import date
from typing import List
from eventkit import Event
from loguru import logger

from optrabot.broker.optionpricedata import OptionStrikePriceData
from optrabot.broker.order import Execution, Order as GenericOrder, OrderStatus
from optrabot.config import Config
from optrabot.models import Account
import optrabot.symbolinfo as symbolInfo
from optrabot.tradetemplate.templatefactory import Template
from optrabot.managedtrade import ManagedTrade

class BrokerConnector(ABC):
	# Constants for Events
	EVENT_CONNECTED = 'connectedEvent'
	EVENT_CONNECT_FAILED = 'connectFailedEvent'
	EVENT_COMMISSION_REPORT = 'commissionReportEvent'
	EVENT_DISCONNECTED = 'disconnectedEvent'
	EVENT_ORDER_STATUS = 'orderStatusEvent'
	EVENT_ORDER_EXEC_DETAILS = 'orderExecutionDetailsEvent'

	def __init__(self) -> None:
		self._initialized = False
		self._createEvents()
		self.id = None		# ID of the connector
		self.broker = None	# ID of the broker
		self._tradingEnabled = False
		self._managedAccounts: List[Account] = []
		pass
	
	@abstractmethod
	async def cancel_order(self, order: GenericOrder):
		"""
		Cancels the given order
		"""
		pass

	@abstractmethod
	async def connect(self):
		""" 
		Establishes a connection to the broker
		"""
		logger.info('Connecting with broker {}', self.id)

	@abstractmethod
	async def disconnect(self):
		""" 
		Disconnects from the broker
		"""
		logger.info('Disconnecting from broker {}', self.id)

	@abstractmethod
	def isConnected(self) -> bool:
		""" 
		Returns True if the broker is connected
		"""
		pass

	@abstractmethod
	def getAccounts(self) -> List[Account]:
		""" 
		Returns the accounts managed by the broker connection
		"""
		pass
	
	@abstractmethod
	async def prepareOrder(self, order: GenericOrder, need_valid_price_data: bool = True) -> bool:
		"""
		Prepares the given order for execution.
		- Retrieve current market data for order legs

		It returns True, if the order could be prepared successfully
		"""

	def oco_as_complex_order(self) -> bool:
		""" 
		Returns True if the broker connection supports OCO orders in form of one complex order
		instead of two single orders
		"""
		return False

	@abstractmethod
	async def placeOrder(self, managed_trade: ManagedTrade, order: GenericOrder, parent_order: GenericOrder = None) -> bool:
		""" 
		Places the given order for a managed account via the broker connection
		"""
		pass

	@abstractmethod
	async def place_complex_order(self, take_profit_order: GenericOrder, stop_loss_order: GenericOrder, template: Template) -> bool:
		""" 
		Places Take Profit and Stop Loss Order as single complex order
		"""
		pass

	@abstractmethod
	async def adjustOrder(self, managed_trade: ManagedTrade, order: GenericOrder, price: float) -> bool:
		""" 
		Adjusts the given order with the given new price
		"""
		pass

	@abstractmethod
	async def requestTickerData(self, symbols: List[str]):
		""" 
		Request ticker data for the given symbols and their options
		"""
		pass

	@abstractmethod
	def getFillPrice(self, order: GenericOrder) -> float:
		""" 
		Returns the fill price of the given order if it is filled
		"""
		pass

	def getLastPrice(self, symbol: str) -> float:
		""" 
		Returns the last price of the given symbol
		"""
		pass

	@abstractmethod
	def get_option_strike_price_data(self, symbol: str, expiration: date, strike: float) -> OptionStrikePriceData:
		""" 
		Returns the option strike price data for the given symbol, expiration, strike and right
		"""
		pass

	@abstractmethod
	def get_strike_by_delta(self, symbol: str, right: str, delta: int) -> float:
		""" 
		Returns the strike price based on the given delta based on the buffered option price data
		"""
		raise NotImplementedError()
	
	def get_min_price_increment(self, price: float) -> float:
		""" 
		Returns the minimum price increment for the given option based on the given price.
		It's based on the information from Tastytrade Help Center article: https://support.tastytrade.com/support/s/solutions/articles/43000435374
		"""
		if price < 3:
			return 0.05
		else:
			return 0.1

	async def eod_settlement_tasks(self):
		"""
		Perform End of Day settlement tasks
		"""
		pass

	def uses_oco_orders(self) -> bool:
		""" 
		Returns True if the broker connection supports and uses OCO orders for managing take profit and stop loss
		"""
		return False

	def _createEvents(self):
		""" 
		Creates the events for the broker connection
		"""
		self.connectedEvent = Event(self.EVENT_CONNECTED)
		self.disconnectedEvent = Event(self.EVENT_DISCONNECTED)
		self.commissionReportEvent = Event(self.EVENT_COMMISSION_REPORT)
		self.connectFailedEvent = Event(self.EVENT_CONNECT_FAILED)
		self.orderStatusEvent = Event(self.EVENT_ORDER_STATUS)
		self.orderExecutionDetailsEvent = Event(self.EVENT_ORDER_EXEC_DETAILS)

	def _emitConnectedEvent(self):
		""" 
		Emits the connected event
		"""
		self.connectedEvent.emit(self)

	def _emitDisconnectedEvent(self):
		""" 
		Emits the disconnected event
		"""
		self._managedAccounts = []
		self.disconnectedEvent.emit(self)

	def _emitConnectFailedEvent(self):
		""" 
		Emits the broker connect failed event
		"""
		self.connectFailedEvent.emit(self)

	def _emitCommissionReportEvent(self, order: GenericOrder, execution_id: str, commission: float = 0, fee: float = 0):
		"""
		Emits the commission report event if commission and fee information are delivered
		for a execution which previously has been reported with the according execution_id.
		"""
		self.commissionReportEvent.emit(order, execution_id, commission, fee)

	def _emitOrderExecutionDetailsEvent(self, order: GenericOrder, execution: Execution):
		""" 
		Emits the order execution details event
		"""
		self.orderExecutionDetailsEvent.emit(order, execution)

	def _emitOrderStatusEvent(self, order: GenericOrder, status: OrderStatus, filledAmount: int = 0):
		""" 
		Emits the order status event. Filled amount holds the amount that has been filled with this
		order status change, if the status event is a "Filled" event.
		"""
		self.orderStatusEvent.emit(order, status, filledAmount)

	def isInitialized(self) -> bool:
		""" 
		Returns True if the broker connector is initialized
		"""
		return self._initialized

	def isTradingEnabled(self) -> bool:
		""" 
		Returns True if trading is enabled
		"""
		return self._tradingEnabled


from typing import Dict
import datetime as dt

class OptionStrikePriceData:
	def __init__(self) -> None:
		self.callBid: float = None
		self.callAsk: float = None
		self.callDelta: float = None
		self.putBid: float = None
		self.putAsk: float = None
		self.putDelta: float = None
		self.brokerSpecific = {}  # Broker specific data
		self.lastUpdated: dt.datetime =None  # Last time the data was updated

	def getPutMidPrice(self) -> float:
		"""
		Returns the mid price of the put option
		"""
		if self.putBid == None:
			bidPrice = 0
		else:
			bidPrice = self.putBid
		
		if self.putAsk == None:
			askPrice = 0
		else:
			askPrice = self.putAsk
		return (bidPrice + askPrice) / 2
	
	def getCallMidPrice(self) -> float:
		"""
		Returns the mid price of the call option
		"""
		if self.callBid == None:
			bidPrice = 0
		else:
			bidPrice = self.callBid
		
		if self.callAsk == None:
			askPrice = 0
		else:
			askPrice = self.callAsk
		return (bidPrice + askPrice) / 2
	
	def is_outdated(self) -> bool:
		"""
		Returns true if the data is outdated
		"""
		if self.lastUpdated == None:
			return True
		delta = dt.datetime.now() - self.lastUpdated
		return delta.total_seconds() > 30

class OptionStrikeData:
	def __init__(self) -> None:
		self.strikeData: Dict[float, OptionStrikePriceData] = {}


import optrabot.config as optrabotcfg
from loguru import logger
from optrabot.broker.brokerfactory import BrokerFactory
from optrabot.broker.order import OptionRight, Order
from optrabot.config import Config
from optrabot.signaldata import SignalData
from optrabot.symbolinfo import symbol_infos
from optrabot.tradetemplate.templatefactory import Template
from optrabot.managedtrade import ManagedTrade

"""
Base class for all template processors
"""
class TemplateProcessorBase:

	def __init__(self, template: Template):
		"""
		Initializes the template processor with the given template
		"""
		self._template = template
		self._config: Config = optrabotcfg.appConfig

	def composeEntryOrder(self, signalData: SignalData = None) -> Order:
		"""
		Composes the entry order based on the template and the optional signal data
		"""
		logger.debug('Creating entry order for template {}', self._template.name)

	def composeTakeProfitOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the take profit order based on the template and the given fill price
		"""
		logger.debug('Creating take profit order for trade {}', managedTrade.trade.id)

	def composeStopLossOrder(self, managedTrade: ManagedTrade, fillPrice: float) -> Order:
		"""
		Composes the stop loss order based on the template and the given fill price
		"""
		logger.debug('Creating stop loss order for trade {}', managedTrade.trade.id)

	def hasTakeProfit(self) -> bool:
		"""
		Returns True if the template has a take profit defined
		"""
		return self._template.hasTakeProfit()
	
	def get_short_strike_from_delta(self, symbol: str, right: OptionRight, delta: int) -> float:
		"""
		Returns the short strike based on the given delta via the associated broker
		connector and the buffered price data
		"""
		brokerConnector = BrokerFactory().getBrokerConnectorByAccount(self._template.account)
		assert brokerConnector != None
		return brokerConnector.get_strike_by_delta(symbol, right, delta)
	
	def check_conditions(self) -> bool:
		"""
		Checks the conditions of the template against the given 
		"""
		if self._template.vix_max:
			logger.debug('Checking Max VIX condition')
			broker = BrokerFactory().getBrokerConnectorByAccount(self._template.account)
			if broker == None:
				logger.error('No broker connection available for account {}', self._template.account)
				return False
			
			try:
				vixPrice = broker.getLastPrice(symbol_infos['VIX'].symbol)
			except Exception as e:
				logger.warning('No price data for VIX available!')
				return False
			logger.debug('VIX Price: {}', vixPrice)
			if vixPrice:
				if vixPrice > self._template.vix_max:
					logger.info(f'Max VIX condition (max: {self._template.vix_max} current: {vixPrice}) not met. Ignoring signal.')
					return False
	
		return True


import asyncio
from enum import Enum
from optrabot import config, crud, schemas
from optrabot.config import Config
import certifi
from contextlib import suppress
import datetime as dt
from datetime import datetime, timedelta
from ib_async import *
import json
from loguru import logger
import httpx
from typing import Any, List
from fastapi import status
from sqlalchemy.orm import Session
import optrabot.config as optrabotcfg
from optrabot.database import get_db_engine
from optrabot.signaldata import SignalData
from optrabot.tradehelper import TradeHelper
from optrabot.optionhelper import OptionHelper
from optrabot.stoplossadjuster import StopLossAdjuster
from optrabot.tradetemplate.processor.templateprocessor import TemplateProcessor
from optrabot.tradetemplate.templatefactory import Template
from optrabot.tradetemplate.templatetrigger import TriggerType
from optrabot.util.singletonmeta import SingletonMeta
from apscheduler.schedulers.asyncio import AsyncIOScheduler
import ssl
import websockets

class NotificationType(str, Enum):
	"""
	Represents the type of an order
	"""
	ERROR = "E"
	INFO = "I"
	WARN = "W"
	

class TradinghubClient(metaclass=SingletonMeta):
	def __init__(self, optraBot):
		logger.debug("TradinghubClient Init")
		self._lastAnswerReceivedAt = None
		self._config = optrabotcfg.appConfig
		self.optraBot = optraBot
		if self.optraBot:
			self._config : Config = self.optraBot['config']
		else:
			self._config = Config()
		self._instanceId = ''
		self.hub_host = ''
		self._contracts: int = 0
		try:
			self._contracts = int(self._config.get('tws.contracts'))
		except KeyError as keyErr:
			self._contracts = 1

		try:
			self.hub_host = self._config.get('general.hub')
		except KeyError as keyError:
			self.hub_host = config.defaultHubHost
			logger.warning("No Hub URL is configured in config.yaml. Using the default.")

		try:
			self._apiKey = self._config.get('general.apikey')
		except KeyError as keyError:
			logger.error("No API Key is configured in config.yaml. Stopping here!")
			self._apiKey = None

		try:
			self._instanceId = self._config.getInstanceId()
			logger.info("Running with Instance ID '{}'.", self._instanceId)
		except KeyError as keyError:
			self._instanceId = None

		self._entryTrade = None
		self._entryTradeContract = None
		self._currentTemplate = None
		self._currentTrade = None
		self._slShortTrade = None
		self._tpTrade = None
		self._ironFlyAskPrice = 0.0
		self._ironFlyComboContract = None
		self._ironFlyContracts = None
		self._ironFlyLongLegContracts = None
		self._ironFlyShortComboContract = None
		self._longLegFillsPrice = 0.0
		self._longLegFillsReceived = 0
		self._previousIronFlyPrice = 0.0
		self._unchangedIronFlyPriceCounter = 0
		self._shuttingdown = False
		self._position = False
		self._closingLongLegs = False
		self._fillTransactionMap = {}
		self._entryOrderFilled = 0   # Number of fills for the entry order
		self._task_web_socket = None
		self._web_socket = None
		self._last_keepalive: datetime = datetime.now()

	async def _monitor_keepalive(self):
		"""
		Monitor the keepalive status and reconnect if necessary
		"""
		while not self._shuttingdown:
			if self._last_keepalive < datetime.now() - timedelta(seconds=30):
				logger.warning('No keep alive received from OptraBot Hub since more than 30 seconds. Reconnecting ...')
				await self._web_socket.close()
			await asyncio.sleep(30)

	async def connect(self, additional_data):
		""" Connects to the OptraBot Hub and reports the livecycle action.
		"""
		logger.info('Connecting to OptraBot Hub ...')
		if not self._use_websocket():
			await self.reportAction(action='SU', additional_data=json.dumps(additional_data))
		else:
			self._task_web_socket = asyncio.create_task(self._connect_websocket(additional_data))

	async def _connect_websocket(self, additional_data):
		"""
		Connects to the OptraBot Hub using a websocket and keeps the connection open. 
		"""
		if self.hub_host.startswith('wss'):
			ssl_context = ssl.create_default_context(cafile=certifi.where())
		else:
			ssl_context = None
	
		keep_alive_task = asyncio.create_task(self._monitor_keepalive())

		while not self._shuttingdown:
			try:
				async with websockets.connect(
					self.hub_host + '/ws', 
					additional_headers=[('X-API-Key', self._apiKey), ('X-Version', self.optraBot.Version)],
					ssl=ssl_context
					) as self._web_socket:
					
					# Send Account Numbers to Hub
					await self._web_socket.send(json.dumps(additional_data))

					# Wait for message that connection has been accepted
					response = await self._web_socket.recv()
					json_data = json.loads(response)
					if json_data.get('message') != 'CONNECTION_ACCEPTED':
						logger.error('Connection to OptraBot Hub failed!')
						return
					
					logger.info('Connected to OptraBot Hub successfully.')

					asyncio.create_task(self._websocket_keepalive())

					while True:
						response = await self._web_socket.recv()
						#logger.debug('Received data from Hub: {}', response)
						try:
							data = json.loads(response)
						except json.JSONDecodeError as jsonExcp:
							logger.error('Received invalid JSON data from Hub: {}', response)

						try:
							action = data['action']
							if action == 'KA':
								#logger.debug('Received Keepalive from Hub.')
								self._last_keepalive = datetime.now()
								continue
						except Exception as excp:
							pass

						try:
							data = json.loads(data)
						except json.JSONDecodeError as jsonExcp:
							logger.error('Received invalid JSON data from Hub: {}', response)
						
						# Check if the recieved data is a signal to be processed
						try:
							signal_data = data["signal"]
							await self._process_signal_data(signal_data)
						except KeyError as keyExcp:
							pass

				logger.info('Connection to OptraBot Hub closed.')

			except websockets.ConnectionClosedError as closedExcp:
				if closedExcp.reason:
					logger.error(f'Connection Closed by OptraBot Hub. Reason: {closedExcp.reason}')
				else:
					logger.error('Connection Closed by OptraBot Hub.')
				if closedExcp.code == 4001:
					break
			except websockets.ConnectionClosedOK as closedExcp:
				logger.info('Connection to OptraBot Hub closed.')
			except Exception as excp:
				logger.error(f'Error connecting to OptraBot Hub: {excp}')
			finally:
				if not self._shuttingdown:
					logger.info('Reconnecting in 10 seconds ...')
					await asyncio.sleep(10)
		keep_alive_task.cancel()

	async def shutdown(self):
		logger.debug('Shutting down Trading Hub Client.')
		self._shuttingdown = True

		# Disconnect from OptraBot Hub
		if not self._use_websocket():
			try:
				logger.info('Disconnecting from OptraBot Hub ...')
				additional_data = {
		 			'instance_id': self._config.getInstanceId()
				}
				await self.reportAction(action='SD', additional_data=json.dumps(additional_data))
			except Exception as excp:
				logger.error('Problem on Shutdown: {}', excp)
		else:
			# Close Websocket connection
			if self._web_socket:
				if self._task_web_socket:
					logger.debug('Stopping Websocket Task.')
					self._task_web_socket.cancel()
				logger.debug('Closing Websocket Connection.')
				try:
					await self._web_socket.send(json.dumps({'action': 'SD', 'instance_id': self._instanceId}))
					await self._web_socket.close()
				except Exception as excp:
					pass
		#self._stopPositionMonitoring()

	async def reportAction(self, action: str, additional_data: str = '') -> bool:
		"""
		Reports a livecycle action and relevant details to the OptraBot Hub and check if it is allowed.
		- Instance with the same id must not be running already
		- A subscription must be active
		- The configured account numbers must not be used by another registered user already

		Returns true if the action was reported successfully, otherwise false.
		"""
		data = {
			'action': action,
			'instance_id': self._instanceId,
			'additional_data': additional_data
		}
		headers = {'X-API-Key': self._apiKey, 'X-Version': self.optraBot.Version, "Content-Type": "application/json"}
		if not self._use_websocket():
			try:
				response = httpx.post(self.hub_host + '/report_action', json=data, headers=headers)
			except Exception as excp:
				raise excp
			
			if response.status_code != status.HTTP_200_OK:
				jsonData = response.json()
				if jsonData['detail']:
					errorMessage = jsonData['detail']
				else:
					errorMessage = response.reason_phrase
				raise Exception(errorMessage)
		else:
			# Send the data via Websocket
			await self._web_socket.send(json.dumps(data))

	async def send_notification(self, notification_type: NotificationType, message: str):
		"""
		Send a Telegram notification via OptraBot Server
		"""
		additional_data = {
		 	'instance_id': self._config.getInstanceId(),
			'type': notification_type,
			'message': message
		}
		await self.reportAction(action='NT', additional_data=json.dumps(additional_data))


	async def _poll(self):
		try:
			fetch_url = self.hub_host + '/fetch_signal'
			url_params = {'instanceid': self._instanceId}
			headers = {'X-API-Key': self._apiKey, 'X-Version': self.optraBot.Version}
			if self._shuttingdown:
				logger.debug("Client Session closed. Stop polling.")
				return
			
			logger.debug('Checking for Signal from Hub.')
			response = httpx.get(fetch_url, params=url_params, follow_redirects=True, headers=headers)
			logger.debug('Answer received ({}).', response.status_code)
			self._pollTimeouts = 0
			if response.status_code != status.HTTP_200_OK:
				if response.status_code == status.HTTP_401_UNAUTHORIZED:
					logger.error("Error requesting signal from OptraBot Hub: {}", response.json()['detail'])
				else:
					logger.error("Error on HTTP request: {}", response.reason_phrase)
				return
			
			self._lastAnswerReceivedAt = datetime.now()
			if response.text != '\"\"' and response.text != '':
				logger.debug("Response {}", response.content )

				try:
					response_data = json.loads(response.content)
				except json.JSONDecodeError as jsonExcp:
					logger.error("Didn't receive JSON data!")
					return
				
				await self._process_signal_data(response_data)
				
		except httpx.TimeoutException as timeoutException:
			logger.debug("Timeout while fetching Signal from Hub: {}", timeoutException)
			self._pollTimeouts += 1
			if self._pollTimeouts > 3:
				logger.error("Too many timeouts while fetching Signal from Hub. Please check the connection!")
				return
		except Exception as anyEcxp:
			logger.error("Exception occured during poll: {}", anyEcxp)

	async def _process_signal_data(self, signal_data: dict):
		"""
		Process signal data received from the OptraBot hub
		"""
		signalStrategy = signal_data['strategy']
		logger.debug('Received Signal: {}', signal_data['strategy'])
		templateProcessor = TemplateProcessor()
		triggeredTemplates = templateProcessor.determineTemplates(signalStrategy)
		if len(triggeredTemplates) == 0:
			logger.warning('No template for external trigger value {} configured!', signal_data['strategy'])
			return
				
		signalData = SignalData(timestamp=self._parseTimestamp(signal_data['time']), close=signal_data['close'], strike=0 )
		
		# Use the VWAP target as strike price for the Iron Fly and the Lunch Iron Fly
		if signalStrategy == '0DTEIronFly' or signalStrategy == '0DTELunchIF':
			signalData.strike = signal_data['vwaptarget']
		else:
			signalData.strike = signal_data['shortstrike']
		
		# Obtain the Template Processor and schedule a job to process the signals
		job_scheduler : AsyncIOScheduler = self.optraBot._backgroundScheduler
		for triggeredTemplate in triggeredTemplates:
			job_id = 'processtemplate' + str(triggeredTemplate.name)
			job_scheduler.add_job(templateProcessor.processTemplate, args=[triggeredTemplate, signalData], id=job_id, max_instances=10, misfire_grace_time=None)
		return	

	# async def excuteTradeTemplate(self, template: Template, response_data: dict):
	# 	""" This method is called via Job Scheduler for executing the given Trade Template in the background.
	# 	"""
	# 	logger.debug(f"Excuting Trade Template {template.name}")
	# 	try:
	# 		ib: IB = self.optraBot['ib']
	# 		if not ib.isConnected():
	# 			logger.error("Interactive Brokers is not connected. Unable to process received signal!")
	# 			return
			
	# 		if not self.optraBot.isTradingEnabled():
	# 			logger.error("Trading is not enabled. Looks like your Market Data Subscription is wrong. Skippimg this Trade!")
	# 			return
			
	# 		spx = Index('SPX', 'SMART')
	# 		qualifiedContracts = await ib.qualifyContractsAsync(spx) 
	# 		[ticker] = await ib.reqTickersAsync(spx)
	# 		spxValue = ticker.marketPrice()
	# 		#self.app['SPXPrice'] = spxValue
	# 		logger.debug("SPX Market Price {}", spxValue)

	# 		chains = await ib.reqSecDefOptParamsAsync(spx.symbol, '', spx.secType, spx.conId)
	# 		chain = next(c for c in chains if c.tradingClass == 'SPXW' and c.exchange == 'SMART')
	# 		if chain == None:
	# 			logger.error("No Option Chain for SPXW and SMART found! Doing no trade!")
	# 			return
			
			
	# 		# Options Kontrakte ermitteln
	# 		nan = float('nan')
	# 		self._currentTemplate = template
	# 		self._currentTemplate.resetStopLossAdjuster()
	# 		if not self._currentTemplate.wing:
	# 			wingSize = 70
	# 		else:
	# 			wingSize = self._currentTemplate.wing
	# 		amount = self._currentTemplate.amount
	# 		accountNo = self._currentTemplate.account
	# 		current_date = dt.date.today()
	# 		expiration = current_date.strftime('%Y%m%d')
	# 		shortLegStrike = float(response_data['vwaptarget'])
	# 		longPutStrike = shortLegStrike - wingSize
	# 		longCallStrike = shortLegStrike + wingSize
	# 		logger.info("Building Iron Fly combo with Short strike {}, Long Put strike {} and Long Call strike {}", shortLegStrike, longPutStrike, longCallStrike)
			
	# 		# Überprüfen das auf den geplanten Legs keine anderen Orders liegen
	# 		logger.debug('Checking for open Trades if Combo Legs are colliding')
	# 		contractsWithOpenOrders = await self._getAllOpenOrderContracts(accountNo, spx.symbol)

	# 		shortPutContract = Option(spx.symbol, expiration, shortLegStrike, 'P', 'SMART', tradingClass = 'SPXW')
	# 		await ib.qualifyContractsAsync(shortPutContract)
	# 		if not OptionHelper.checkContractIsQualified(shortPutContract):
	# 			return
	# 		if shortPutContract.conId in contractsWithOpenOrders:
	# 			logger.error('Existing open order at stike price {}. Trade cannot be placed!', shortLegStrike)
	# 			return

	# 		shortCallContract = Option(spx.symbol, expiration, shortLegStrike, 'C', 'SMART', tradingClass = 'SPXW')				
	# 		await ib.qualifyContractsAsync(shortCallContract)
	# 		if not OptionHelper.checkContractIsQualified(shortCallContract):
	# 			return
	# 		if shortCallContract.conId in contractsWithOpenOrders:
	# 			logger.error('Existing open order at strike price {}. Trade cannot be placed!', shortLegStrike)
	# 			return
			
	# 		longPutContract = await self._DetermineValidLongOption(spx.symbol, expiration, longPutStrike, 'P', contractsWithOpenOrders)
	# 		if longPutContract == None or not OptionHelper.checkContractIsQualified(longPutContract):
	# 			logger.error('Unable to determine a valid Long Put option for the trade!')
	# 			return
	# 		if longPutContract.strike != longPutStrike:
	# 			logger.warning('Using different Long Put strike {}, because of existing open orders on desired strike {}.', longPutContract.strike, longPutStrike)

	# 		longCallContract = await self._DetermineValidLongOption(spx.symbol, expiration, longCallStrike, 'C', contractsWithOpenOrders)
	# 		if longCallContract == None or not OptionHelper.checkContractIsQualified(longCallContract):
	# 			logger.error('Unable to determine a valid Long Call option for the trade!')
	# 			return
	# 		if longCallContract.strike != longCallStrike:
	# 			logger.warning('Using different Long Call strike {}, because of existing open orders on desired strike {}.', longCallContract.strike, longCallStrike)

	# 		self._ironFlyLongLegContracts = [longCallContract, longPutContract]
	# 		self._ironFlyContracts = [shortPutContract, shortCallContract, longPutContract, longCallContract]
	# 		self._ironFlyComboContract = Contract(symbol=spx.symbol, secType='BAG', exchange='SMART', currency='USD', comboLegs=[])
	# 		self._ironFlyShortComboContract = Contract(symbol=spx.symbol, secType='BAG', exchange='SMART', currency='USD', comboLegs=[])
	# 		ironFlyMidPrice = None
	# 		for i in range(5):
	# 			tickers = await ib.reqTickersAsync(*self._ironFlyContracts)
	# 			logger.debug("Tickers {}", tickers)
	# 			ironFlyMidPrice = self._calculateIronFlyMidPrice(tickers, ironFlyComboContract=self._ironFlyComboContract, 
	# 										   ironFlyShortComboContract=self._ironFlyShortComboContract, 
	# 										   shortPutContract=shortPutContract,
	# 										   shortCallContract=shortCallContract,
	# 										   longPutContract=longPutContract,
	# 										   longCallContract=longCallContract)
	# 			if ironFlyMidPrice != None:
	# 				break
	# 			await asyncio.sleep(1)
			
	# 		#if util.isNan(ironFlyMidPrice):
	# 		if ironFlyMidPrice == None:
	# 			logger.error("No Mid Price for combo could be calculated!")
	# 			return

	# 		limitPrice = ironFlyMidPrice

	# 		logger.info("IronFly Combo Mid Price: {} Ask Price: {}", ironFlyMidPrice, self._ironFlyAskPrice)

	# 		if not self._meetsMinimumPremium(limitPrice):
	# 			logger.info('Premium below configured minimum premium of ${}. Trade is not executed!', self._currentTemplate.minPremium)
	# 			return
			
	# 		order = LimitOrder('BUY', amount, limitPrice)
	# 		with Session(get_db_engine()) as session:
	# 			newTrade = schemas.TradeCreate(account=accountNo, symbol='SPX', strategy=self._currentTemplate.strategy)
	# 			self._currentTrade = crud.create_trade(session, newTrade)
	# 			order.orderRef = 'OTB (' + str(self._currentTrade.id) + '): ' + template.name + ' - Open'
	# 		order.account = accountNo
	# 		order.outsideRth = True
	# 		self._fillTransactionMap.clear()
	# 		self._entryTrade = ib.placeOrder(self._ironFlyComboContract, order)
	# 		self._entryTradeContract = self._ironFlyComboContract
	# 		self._ironFlyAskPrice = 0.0
	# 		self._longLegFillsPrice = 0.0
	# 		self._previousIronFlyPrice = 0.0
	# 		self._unchangedIronFlyPriceCounter = 0
	# 		self._longLegFillsReceived = 0
	# 		self._slShortTrade = None
	# 		self._tpTrade = None
	# 		self._entryOrderFilled = 0
	# 		self._closingLongLegs = False
	# 		logger.debug("Account: {} Trade placed: {} Number of contracts: {}", order.account, self._entryTrade, amount)
	# 		self.optraBot._backgroundScheduler.add_job(self._trackEntryOrder, 'interval', seconds=5, id='TrackEntryOrder')

	# 	except Exception as excp:
	# 					logger.error("Exception: {}", excp)

	# def _calculateIronFlyMidPrice(self, tickers: List[Ticker], ironFlyComboContract: Contract, ironFlyShortComboContract: Contract, shortPutContract: Option, shortCallContract: Option, longPutContract: Option, longCallContract: Option) -> float:
	# 	"""
	# 	Calculate the Iron Fly Mid Price based on the given tickers. It returns
	# 	None if one of the legs got no valid price.
	# 	"""
	# 	ironFlyMidPrice = 0
	# 	self._ironFlyAskPrice = 0
	# 	ironFlyComboContract.comboLegs.clear()
	# 	ironFlyShortComboContract.comboLegs.clear()
	# 	for ticker in tickers:
	# 		tickerContract = ticker.contract
	# 		if tickerContract.conId == shortPutContract.conId:
	# 			midPrice = (ticker.ask + ticker.bid) / 2
	# 			if util.isNan(midPrice) or (ticker.ask == -1.00 and ticker.bid == -1.00):
	# 				return None
	# 			ironFlyMidPrice -= midPrice
	# 			if not util.isNan(ticker.bid):
	# 				self._ironFlyAskPrice -= ticker.bid
	# 				ironFlyComboContract.comboLegs.append(ComboLeg(conId=shortPutContract.conId, ratio=1, action='SELL', exchange='CBOE'))
	# 				ironFlyShortComboContract.comboLegs.append(ComboLeg(conId=shortPutContract.conId, ratio=1, action='BUY', exchange='CBOE'))
	# 		if tickerContract.conId == shortCallContract.conId:
	# 			midPrice = (ticker.ask + ticker.bid) / 2
	# 			if util.isNan(midPrice) or (ticker.ask == -1.00 and ticker.bid == -1.00):
	# 				return None
	# 			ironFlyMidPrice -= midPrice
	# 			if not util.isNan(ticker.bid):
	# 				self._ironFlyAskPrice -= ticker.bid
	# 				ironFlyComboContract.comboLegs.append(ComboLeg(conId=shortCallContract.conId, ratio=1, action='SELL', exchange='CBOE'))
	# 				ironFlyShortComboContract.comboLegs.append(ComboLeg(conId=shortCallContract.conId, ratio=1, action='BUY', exchange='CBOE'))
	# 		if tickerContract.conId == longPutContract.conId:
	# 			midPrice = (ticker.ask + ticker.bid) / 2
	# 			if util.isNan(midPrice) or (ticker.ask == -1.00 and ticker.bid == -1.00):
	# 				midPrice = 0.05
	# 			ironFlyMidPrice += midPrice
	# 			if not util.isNan(ticker.ask):
	# 				self._ironFlyAskPrice += ticker.ask
	# 				ironFlyComboContract.comboLegs.append(ComboLeg(conId=longPutContract.conId, ratio=1, action='BUY', exchange='CBOE'))
	# 		if tickerContract.conId == longCallContract.conId:
	# 			midPrice = (ticker.ask + ticker.bid) / 2
	# 			if util.isNan(midPrice) or (ticker.ask == -1.00 and ticker.bid == -1.00):
	# 				midPrice = 0.05
	# 			ironFlyMidPrice += midPrice
	# 			if not util.isNan(ticker.ask):
	# 				self._ironFlyAskPrice += ticker.ask
	# 				ironFlyComboContract.comboLegs.append(ComboLeg(conId=longCallContract.conId, ratio=1, action='BUY', exchange='CBOE'))
	# 	self._ironFlyAskPrice = OptionHelper.roundToTickSize(self._ironFlyAskPrice)
	# 	return OptionHelper.roundToTickSize(ironFlyMidPrice)

	# async def _trackEntryOrder(self):
	# 	ib: IB = self.optraBot['ib']
	# 	if self._entryTrade == None:
	# 		logger.debug("Stopping Entry Order Tracking")
	# 		self.optraBot._backgroundScheduler.remove_job('TrackEntryOrder')
	# 		return
		
	# 	if self._entryTrade.orderStatus.status == OrderStatus.Cancelled or self._entryTrade.orderStatus.status == OrderStatus.Inactive:
	# 		self._onEntryOrderCanceled()
	# 	elif self._entryTrade.orderStatus.status == OrderStatus.Filled:
	# 		logger.info("Entry Order has been filled already. No adjustment required")
	# 		logger.debug("Stopping Entry Order Tracking")
	# 		self.optraBot._backgroundScheduler.remove_job('TrackEntryOrder')
	# 	elif self._entryTrade.orderStatus.status == OrderStatus.PreSubmitted or self._entryTrade.orderStatus.status == OrderStatus.PendingSubmit:
	# 		logger.info("Entry Order is still in PreSubmitted state. Waiting for it to be submitted...")
	# 	elif self._entryTrade.orderStatus.status == OrderStatus.Submitted:
	# 		currentLimitPrice = self._entryTrade.order.lmtPrice
	# 		adjustedLimitPrice = currentLimitPrice + self._currentTemplate.adjustmentStep
	# 		logger.info("Entry Order status ({}). Entry price will be adjusted. Current Limit Price: ${}", self._entryTrade.orderStatus.status, currentLimitPrice)
	# 		if self._meetsMinimumPremium(adjustedLimitPrice) and adjustedLimitPrice <= self._ironFlyAskPrice:
	# 			self._entryTrade.order.lmtPrice = adjustedLimitPrice
	# 			try:
	# 				ib.placeOrder(self._entryTradeContract, self._entryTrade.order)
	# 			except Exception as excp:
	# 				logger('Exception beim Anpassen der Order')
	# 			return
	# 		else:
	# 			if adjustedLimitPrice > -15:
	# 				logger.info("Entry order limit price reached minimum premium. No entry.")
	# 			if adjustedLimitPrice > self._ironFlyAskPrice:
	# 				logger.info("Entry order limit price exceeded initial ask price. No entry.")
	# 			ib.cancelOrder(self._entryTrade.order)
	# 	else:
	# 		logger.error("Entry Order status is not considered: {}", self._entryTrade.orderStatus.status)

	# async def onOrderStatusEvent(self, trade: Trade):
	# 	if trade == self._entryTrade:
	# 		logger.debug('Order Status Event has been raised. Status: {}', trade.orderStatus.status)
	# 		if trade.orderStatus.status == OrderStatus.Cancelled:
	# 			logEntry: TradeLogEntry = None
	# 			for logEntry in trade.log:
	# 				if logEntry.status != 'Cancelled':
	# 					continue
	# 			if logEntry == None:
	# 				logger.error('No log entry found for Entry order cancellation!')
	# 			if logEntry.errorCode == 103:
	# 				# Error 103, reqId 10292: Doppelt vorhandene Order-ID
	# 				logger.info('Adjustment of entry order has been rejected, because Duplicate Order-ID. Entry Order still active.')
	# 			elif logEntry.errorCode == 0:
	# 				self._onEntryOrderCanceled()
	# 		elif trade.orderStatus.status == OrderStatus.Filled:
	# 			if self._position == False:
	# 				logger.info('Entry Order has been filled at ${} (Qty: {}) and trade is running now.', trade.orderStatus.avgFillPrice, trade.orderStatus.filled)
	# 				self._position = True
	# 				if self._currentTemplate.stopLossAdjuster:
	# 					self._currentTemplate.stopLossAdjuster.setBasePrice(round(self._entryTrade.orderStatus.avgFillPrice * -1, 2))
	# 				asyncio.create_task(self._placeTakeProfitAndStop(trade)).add_done_callback(self.optraBot.handleTaskDone)
	# 				#Run 1. Position Monitoring with delay
	# 				#asyncio.create_task(self._monitorPositionDelayed()).add_done_callback(self.optraBot.handleTaskDone)
	# 				self.optraBot._backgroundScheduler.add_job(self._monitorPosition, 'interval', seconds=5, id='MonitorPosition')
	# 			elif self._position == True:
	# 				logger.debug('Additionally fill quantity (Qty: {}) for the entry order...depending orders need to be adjusted.', trade.orderStatus.filled)
	# 				asyncio.create_task(self._adjustTakeProfitAndStop(trade)).add_done_callback(self.optraBot.handleTaskDone)

	# 			if trade.orderStatus.remaining > 0:
	# 				logger.warning('Partial fill for entry order. Remaining: {}', trade.orderStatus.remaining)
	# 			executedAmount = trade.orderStatus.filled - self._entryOrderFilled
	# 			logger.debug('Executed amount for entry trade: {}', executedAmount)
	# 			if executedAmount > 0:
	# 				asyncio.create_task(self._reportExecutedTrade(trade, executedAmount)).add_done_callback(self.optraBot.handleTaskDone)
	# 			self._entryOrderFilled = self._entryTrade.order.totalQuantity - trade.orderStatus.remaining

	# 	elif self._tpTrade and trade.order.orderId == self._tpTrade.order.orderId:
	# 		logger.debug('TP Order Status has been raised. Status: {}', trade.orderStatus.status)
	# 		if trade.orderStatus.status == OrderStatus.Cancelled:
	# 			logger.info('TP Order has been cancelled!')

	# 		elif trade.orderStatus.status == OrderStatus.Filled:
	# 			logger.success('TP Order has been filled. Trade finished')
	# 			self._tpTrade = None
	# 			self._slShortTrade = None
	# 			self._onPositionClose()

	# 	elif self._slShortTrade and trade.order.orderId == self._slShortTrade.order.orderId:
	# 		logger.debug('SL order for Short Legs status has been changed. Status: {}', trade.orderStatus.status)
	# 		if trade.orderStatus.status == OrderStatus.Cancelled:
	# 			logEntry: TradeLogEntry = None
	# 			for logEntry in trade.log:
	# 				if logEntry.status != 'Cancelled':
	# 					continue
	# 			if logEntry == None:
	# 				logger.error('No log entry found for SL order cancellation!')
	# 				return
	# 			if logEntry.errorCode == 201:
	# 				# 201: Order rejected - reason:Stop price revision is disallowed after order has triggered
	# 				logger.info('Adjustment of SL order has been rejected, because Stop Order was triggered or cancelled already.')
	# 			elif logEntry.errorCode == 0:
	# 				logger.info('SL Order for Short Legs has been cancelled.')
	# 			else:
	# 				logger.error('SL order for Short legs has been cancelled with error code {} and message: {}', logEntry.errorCode, logEntry.message)					

	# 		elif trade.orderStatus.status == OrderStatus.Filled:
	# 			logger.info('SL order for Short Legs has been filled. Trade finished')
	# 			logger.info('Now....Long Legs need to be closed if possible')
	# 			if self._closingLongLegs == False:
	# 				asyncio.create_task(self._close_long_legs(self._entryTrade.orderStatus.filled, self._currentTemplate.name, self._currentTemplate.account, tradeId=self._currentTrade.id )).add_done_callback(self.optraBot.handleTaskDone)
	# 				self._onPositionClose()

	# def _onEntryOrderCanceled(self):
	# 	""" Performs final steps after a entry order was canceled
	# 	"""
	# 	logger.info('Entry Order has been cancelled!')
	# 	self._entryTrade = None
	# 	logger.debug('Deleting trade {} from database...', self._currentTrade.id )
	# 	with Session(get_db_engine()) as session:
	# 		crud.delete_trade(session, self._currentTrade)
	# 	self._currentTrade = None

	# async def _DetermineValidLongOption(self, symbol, expiration, desiredStrike, right, contractsWithOpenOrders):
	# 	""" Determines the next possible Long Leg Option on the given Strike price and right ensuring that
	# 		there are no open orders.
	# 	"""
	# 	ib: IB = self.optraBot['ib']
	# 	strike = desiredStrike
	# 	validOption: Option = None
	# 	tryCount = 0
	# 	while tryCount < 5:
	# 		tryCount += 1
	# 		logger.debug('Checking Long {} Strike {}', right, strike)
	# 		option = Option(symbol, expiration, strike, right, 'CBOE', tradingClass = 'SPXW')
	# 		await ib.qualifyContractsAsync(option)
	# 		if not OptionHelper.checkContractIsQualified(option):
	# 			break
	# 		if not option.conId in contractsWithOpenOrders:
	# 			validOption = option
	# 			break
	# 		if right == 'P':
	# 			strike -= 5
	# 		else:
	# 			strike += 5
	# 	return validOption


	# async def onExecDetailsEvent(self, trade: Trade, fill: Fill):
	# 	""" This eventhandler is called on trade execution
	# 	"""
	# 	# First we need to check if Execution is related to an OptraBot Trade
	# 	# If so, the transaction needs to be recorded in the database
	# 	logger.debug('Exec Detail for trade {}', trade)
	# 	logger.debug('Fill: {}', fill)
	# 	tradeId = TradeHelper.getTradeIdFromOrderRef(fill.execution.orderRef)
	# 	if tradeId == 0:
	# 		logger.debug('Fill not from OptraBot trade...ignoring it.')
	# 		return
	# 	if fill.contract.secType == 'BAG':
	# 		# Do not store information on combo executions
	# 		return
	# 	with Session(get_db_engine()) as session:
	# 		max_transactionid = crud.getMaxTransactionId(session, tradeId)
	# 		dbTrade = crud.getTrade(session,tradeId)
	# 		if max_transactionid == 0:
	# 			# Opening Transaction of the trade
	# 			dbTrade.status = 'OPEN'
	# 		max_transactionid += 1
	# 		transactionType = ''
	# 		if fill.execution.side == 'BOT':
	# 			transactionType = 'BUY'
	# 		else:
	# 			transactionType = 'SELL'
	# 		expirationDate = datetime.strptime(fill.contract.lastTradeDateOrContractMonth, '%Y%m%d')
	# 		newTransaction = schemas.TransactionCreate(tradeid=tradeId, id=max_transactionid, type=transactionType, sectype=fill.contract.right, timestamp=fill.execution.time, expiration=expirationDate, strike=fill.contract.strike, contracts=fill.execution.shares, price=fill.execution.avgPrice, fee=0, commission=0, notes='')
	# 		self._fillTransactionMap[fill.execution.execId] = newTransaction.id # Memorize Execution ID of the fill for later commission report
	# 		crud.createTransaction(session, newTransaction)

	# 		# Check if trade is closed with all these transactions
	# 		TradeHelper.updateTrade(dbTrade)
	# 		session.commit()
	
	async def onCommissionReportEvent(self, trade: Trade, fill: Fill, report: CommissionReport):
		"""
		Handles the Commission Report Event
		"""
		logger.debug('Commission Report order id {} and fill {}', trade.order.orderId, fill)
		logger.debug('Commission Report: {}', report)
		tradeId = TradeHelper.getTradeIdFromOrderRef(fill.execution.orderRef)
		if tradeId == 0:
			logger.debug('Fill not from OptraBot trade...ignoring it.')
			return
		
		# Get Transaction ID for the fills execution id
		try:
			transactionId = self._fillTransactionMap[fill.execution.execId]
		except KeyError as keyError:
			logger.error('No trade transaction found for fill execution id {}', fill.execution.execId)
			return
		
		with Session(get_db_engine()) as session:
			transaction = crud.getTransactionById(session, tradeId, transactionId)
			if transaction == None:
				logger.error('Transaction with id {} for trade {} not found in database!', transactionId, tradeId)
				return
			transaction.commission = report.commission
			session.commit()

	async def _reportExecutedTrade(self, trade: Trade, executedAmount: int):
		"""
		Report the entry trade execution to Optrabot Hub
		"""
		try:
			executedContracts = 0
			if trade.contract.secType == 'BAG':
				for leg in trade.contract.comboLegs:
					executedContracts += leg.ratio * executedAmount
			else:
				executedContracts = executedAmount
					
			additional_data = {
				'trade_id': self._currentTrade.id,
				'account': self._currentTemplate.account,
				'contracts': int(executedContracts)
			}
			await self.reportAction('CT', additional_data=json.dumps(additional_data))
		except Exception as excp:
			logger.error('Error reporting position open event: {}', excp)

	async def _placeTakeProfitAndStop(self, entryTrade: Trade):
		fillPrice = entryTrade.orderStatus.avgFillPrice
		fillAmount = entryTrade.orderStatus.filled
		logger.debug('Entry Order was filled at ${}', fillPrice)
		if entryTrade.orderStatus.remaining > 0:
			logger.warning('Entry Order was filled partially only.')
		ib: IB = self.optraBot['ib']

		# Calculate Take Profit price
		logger.debug('Using Profit Level {} and Stop Level {}', self._currentTemplate.takeProfit, self._currentTemplate.stopLoss)
		profit = fillPrice * (self._currentTemplate.takeProfit / 100)

		limitPrice = OptionHelper.roundToTickSize(fillPrice - profit)
		now = datetime.now()
		ocaGroup = self._currentTemplate.account + '_' + now.strftime('%H%M%S')
		logger.info('Take Profit Limit Price ({}%) : {}', self._currentTemplate.takeProfit, limitPrice)
		order = LimitOrder('SELL', fillAmount, limitPrice)
		order.account = self._currentTemplate.account
		order.orderRef = 'OTB (' + str(self._currentTrade.id) + '): ' + self._currentTemplate.name + ' - Take Profit'
		order.ocaGroup = ocaGroup
		order.outsideRth = True
		self._tpTrade = ib.placeOrder(self._ironFlyComboContract, order)
		logger.debug('Created TP order with id: {}', self._tpTrade.order.orderId)

		# Calculation of Stop Price
		stopFactor = ((self._currentTemplate.stopLoss / 100) + 1) * -1
		self._ironFlyStopPrice = OptionHelper.roundToTickSize(fillPrice * stopFactor)
		logger.info('Stop LossPrice ({}%): {}', self._currentTemplate.stopLoss, self._ironFlyStopPrice)

		stopShortsOrder = StopOrder('BUY', self._entryTrade.orderStatus.filled, self._ironFlyStopPrice)
		stopShortsOrder.account = self._currentTemplate.account
		stopShortsOrder.orderRef = 'OTB (' + str(self._currentTrade.id) + '): ' + self._currentTemplate.name + ' - SL Short Legs'
		stopShortsOrder.ocaGroup = ocaGroup
		try:
			self._slShortTrade = ib.placeOrder(self._ironFlyShortComboContract, stopShortsOrder)
		except Exception as excp:
			logger('Exception beim Erstellen der Short Leg Stoploss order')

	async def _adjustTakeProfitAndStop(self, trade: Trade):
		""" Adjust Take Profit and Stop Order after another partial fill of the entry Order.
		"""
		ib: IB = self.optraBot['ib']
		newQuantity = trade.orderStatus.filled
		if newQuantity == self._tpTrade.order.totalQuantity:
			return
		logger.info('Adjusting quantity of TP and Stop Order to {}', newQuantity)
		self._tpTrade.order.totalQuantity = trade.orderStatus.filled
		ib.placeOrder(self._tpTrade.contract, self._tpTrade.order)
		logger.debug('Updated TP order with id: {}', self._tpTrade.order.orderId)
		self._slShortTrade.order.totalQuantity = trade.orderStatus.filled
		ib.placeOrder(self._slShortTrade.contract, self._slShortTrade.order)
		logger.debug('Updated SL order with id: {}', self._slShortTrade.order.orderId)

	async def _close_long_legs(self, positionSize: int, templatename: str, account: str, tradeId: int):
		logger.debug("Closing long legs of trade")
		self._closingLongLegs = True
		ib: IB = self.optraBot['ib']
		tickers = await ib.reqTickersAsync(*self._ironFlyLongLegContracts)
		for ticker in tickers:
			logger.debug("Long Leg {} {} Bid Price: {}", ticker.contract.right, ticker.contract.strike, ticker.bid)
			if ticker.bid >= 0.05:
				logger.debug('Creating limit sell order on bid price')
				order = LimitOrder('SELL', positionSize, ticker.bid)
				order.outsideRth = True
				order.account = account
				order.orderRef = 'OTB (' + str(tradeId) + '): ' + templatename + ' - Close Long Leg'
				closeTrade = ib.placeOrder(ticker.contract, order)
				logger.debug('Placed Long Leg closing order with id: {}', closeTrade.order.orderId)
			else:
				logger.info('Long leg {} {} is worthless and will be kept open till expiration.', ticker.contract.right, ticker.contract.strike)
				with Session(get_db_engine()) as session:
					max_transactionid = crud.getMaxTransactionId(session, tradeId)
					dbTrade = crud.getTrade(session,tradeId)
					assert(max_transactionid > 0)
					# Create expiration transaction for this leg
					max_transactionid += 1
					transactionType = 'EXP'
					expirationDate = datetime.strptime(ticker.contract.lastTradeDateOrContractMonth, '%Y%m%d')
					newTransaction = schemas.TransactionCreate(tradeid=tradeId, id=max_transactionid, type=transactionType, sectype=ticker.contract.right, timestamp=expirationDate, expiration=expirationDate, strike=ticker.contract.strike, contracts=positionSize, price=0.0, fee=0, commission=0.0, notes='')
					crud.createTransaction(session, newTransaction)
					TradeHelper.updateTrade(dbTrade)
					session.commit()

		logger.debug('exit _close_long_legs()')

	def _onPositionClose(self):
		logger.debug('_onPositionClose()')
		self._tpTrade = None
		self._slShortTrade = None
		self._entryTrade = None
		self._ironFlyComboContract = None
		self._ironFlyContracts = None
		self._position = False
		self._closingLongLegs = False
		self._currentTemplate = None
		self._currentTrade = None
		self._stopPositionMonitoring()

	async def _monitorPosition(self):
		logger.debug('Enter Monitor position()')
		if self._position == False:
			logger.debug('Position has been closed. Stopping Position-Monitoring now.')
			#self._positionMonitorTask = None
			self._stopPositionMonitoring()
			return

		#asyncio.create_task(self._monitorPositionDelayed()).add_done_callback(self.optraBot.handleTaskDone)

		ib: IB = self.optraBot['ib']
		if not ib.isConnected() or self.optraBot.isTradingEnabled() == False:
			logger.error('Interactive Brokers is not connected. Unable to monitor position!')
			return
		
		try:
			if self._position == False:
				logger.debug('Position has been closed. Stopping Position-Monitoring now.')
				return

			tickers = await ib.reqTickersAsync(*self._ironFlyContracts)
			longLegsValue = 0
			currentIronFlyAskPrice = 0
			currentIronFlyBidPrice = 0
			adjustedStopLossPrice = None
			for ticker in tickers:
				if ticker.contract in self._ironFlyLongLegContracts:
					if ticker.bid >= 0:
						longLegsValue += ticker.bid
						currentIronFlyBidPrice -= ticker.bid
						currentIronFlyAskPrice += ticker.ask
				else:
					# Für Short Legs müssen die Ask Preise addiert werden
					currentIronFlyBidPrice += ticker.ask
					currentIronFlyAskPrice -= ticker.bid
		
			if self._position == False:
				logger.debug('Position has been closed. Stopping Position-Monitoring now.')
				return

			currentIronFlyAskPrice = abs(currentIronFlyAskPrice)
			currentIronFlyPrice = OptionHelper.calculateMidPrice(currentIronFlyBidPrice, currentIronFlyAskPrice)
			if currentIronFlyPrice == self._previousIronFlyPrice:
				self._unchangedIronFlyPriceCounter += 1
			else:
				self._unchangedIronFlyPriceCounter = 0
			self._previousIronFlyPrice = currentIronFlyPrice
			if self._unchangedIronFlyPriceCounter > 10:
				logger.warning('Price of Iron Fly has not changed for 10 times. There is a problem with price data from TWS probably.')
				self._unchangedIronFlyPriceCounter = 0

			if self._currentTemplate.stopLossAdjuster:
				if not self._currentTemplate.stopLossAdjuster.isTriggered():
					adjustedStopLossPrice = self._currentTemplate.stopLossAdjuster.execute(currentIronFlyPrice)
					if adjustedStopLossPrice == None:
						logger.debug('No need to adjust stop loss now.')
					else:
						self._ironFlyStopPrice = adjustedStopLossPrice
						logger.info('Performing stop loss adjustment. New Stop Loss price: {}', self._ironFlyStopPrice)
				else:
					logger.debug('Stoploss adjustment has been performed already.')
			else:
				logger.debug('No StopLoss adjustment configured.')
			desiredStopPrice = OptionHelper.roundToTickSize(self._ironFlyStopPrice + longLegsValue)
			currentStopPrice = OptionHelper.roundToTickSize(self._slShortTrade.order.auxPrice)
			logger.debug('Long Legs value {} Current Short SL Price: {} Desired Short SL Pice: {}', round(longLegsValue,2), currentStopPrice, desiredStopPrice)
			openOrders = ib.openOrders()

			if self._position == False:
				logger.debug('Position has been closed. Stopping Position-Monitoring now.')
				return
			if any(order.orderId == self._slShortTrade.order.orderId for order in openOrders):
			#if self._slShortTrade.order in openOrders:
				if self._slShortTrade.isActive():
					if currentStopPrice != desiredStopPrice:
						self._slShortTrade.order.auxPrice = desiredStopPrice
						logger.info('Adjusting Stop Loss price to ${}', desiredStopPrice)
						ib.placeOrder(self._slShortTrade.contract, self._slShortTrade.order)
						logger.debug('Updated SL order with id: {}', self._slShortTrade.order.orderId)
					else:
						logger.debug('No adjustment of Stop Loss price required.')
				else:
					logger.debug('SL Order is not active anymore.')
			else:
				logger.warning('Caution: Stop Loss order for Short Strikes is missing. Trying to reestablish the Stop Loss order now!')
				await self._reestablishStopLossOrder(desiredStopPrice)
		except ConnectionError as connError:
			logger.error('Connection to TWS lost during Position Monitoring. Unable to monitor position!')

		logger.debug('Leave Monitor position()')

	async def _monitorPositionDelayed(self):
		logger.debug('Waiting 10 seconds for next position monitoring.')
		await asyncio.sleep(10)
		if self._position == True:
			asyncio.create_task(self._monitorPosition(), name='MonitorPosition').add_done_callback(self.optraBot.handleTaskDone)

	async def _reestablishStopLossOrder(self, desiredStopPrice: float):
		"""
		Tries to create a new Stop Loss order for the open position based on the given stop price.
		"""
		logger.debug('Reestablishing Stop Loss order for Short Legs')
		ib: IB = self.optraBot['ib']
		if not ib.isConnected():
			logger.error('Interactive Brokers is not connected. Unable to reestablish Stop Loss Order!')
			return
		
		now = datetime.now()
		ocaGroup = self._currentTemplate.account + '_' + now.strftime('%H%M%S')
		stopShortsOrder = StopOrder('BUY', self._entryTrade.orderStatus.filled, desiredStopPrice)
		stopShortsOrder.account = self._currentTemplate.account
		stopShortsOrder.orderRef = 'OTB (' + str(self._currentTrade.id) + '): ' + self._currentTemplate.name + ' - SL Short Legs'
		stopShortsOrder.ocaGroup = ocaGroup
		try:
			self._slShortTrade = ib.placeOrder(self._ironFlyShortComboContract, stopShortsOrder)
		except Exception as excp:
			logger('Exception beim Erstellen der Short Leg Stoploss order')
		logger.info('Stop Loss order for Short Legs has been reestablished.')

		# Check if Take Profit Order is still in place and recreate it, if it's missing
		openTrades = await ib.reqOpenOrdersAsync()
		if self._tpTrade in openTrades:
			logger.info('Take Profit order is still in place.')
		else:
			logger.warning('Take Profit order is missing. Trying to reestablish the Take Profit order now!')
			order = LimitOrder('SELL', self._tpTrade.order.totalQuantity, self._tpTrade.order.lmtPrice)
			order.account = self._currentTemplate.account
			order.orderRef = 'OTB (' + str(self._currentTrade.id) + '): ' + self._currentTemplate.name + ' - Take Profit'
			order.ocaGroup = ocaGroup
			order.outsideRth = True
			self._tpTrade = ib.placeOrder(self._ironFlyComboContract, order)
			logger.debug('Created TP order with id: {}', self._tpTrade.order.orderId)

	def _meetsMinimumPremium(self, premium: float) -> bool:
		""" Checks if given premium meets minimum premium

		Parameters
		----------
		premium : float
			As premium is a typically a credit, a negative number is expected.
		
		Returns
		-------
		bool
			Returns True, if the received premium is more than the configured minimum premium
		"""
		if self._currentTemplate.minPremium == None:
			return True
		if premium > (self._currentTemplate.minPremium * -1):
			return False
		return True

	def _parseTimestamp(self, timestamp: str) -> datetime:
		""" Parses the given timestamp into a `datetime`object

		Parameters
		----------
		timestamp : str
    		Timestamp as string with timezone info e.g. 2023-11-07T14:10:00Z.
		"""
		try:
			parsedTime = datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S%z')
		except Exception as excpt:
			logger.error("Timestamp {} got unexpected format.", timestamp)
			return None
		return parsedTime

	async def _scheduleNextPoll(self):
		await asyncio.sleep(5)
		asyncio.create_task(self._poll()).add_done_callback(self.optraBot.handleTaskDone)
		
	def start_polling(self, scheduler):
		"""
		Tradinghub Polling runner
		"""
		logger.debug("start_polling")
		if self._use_websocket():
			logger.debug("Websocket connection is used. Stop polling.")
			return
		else:
			logger.warning("You're using legacy long polling connection to OptraBot Hub. Please consider using Websocket connection.")
		
		if self._apiKey == None:
			logger.debug("No API Key configured. Stop polling.")
			return

		if self._instanceId == None:
			logger.error("No Instance ID configured in config.yaml. Stop polling!")
			return

		scheduler.add_job(self._poll, 'interval', seconds=5)

	def _stopPositionMonitoring(self):
		self.optraBot._backgroundScheduler.remove_job('MonitorPosition')
	
	def _use_websocket(self):
		""" Returns True if the OptraBot Hub should be accessed via Websocket
		"""
		if self.hub_host.startswith('ws'):
			return True
		return False
	
	async def _websocket_keepalive(self):
		"""
		Asynchronous Task sending a keepalive message to the OptraBot Hub every 20 seconds
		"""
		try:
			while self._web_socket:
				await asyncio.sleep(20)
				await self._web_socket.send('{"action":"KA"}')
				self._lastAnswerReceivedAt = datetime.now()
		except websockets.ConnectionClosedError as closedError:
			pass
		except websockets.ConnectionClosedOK as closedError:
			pass
		logger.error("Websocket connection is not established. Stop keepalive.")

	def isHubConnectionOK(self) -> bool:
		""" Returns True if the last request to the OptraBot Hub was responed 
		30 seconds ago or less.
		"""
		if self._lastAnswerReceivedAt == None or self._last_keepalive == None:
			return False
		timeDelta_ka_send = datetime.now() - self._lastAnswerReceivedAt
		timeDelta_ka_received = datetime.now() - self._last_keepalive
		if timeDelta_ka_send.total_seconds() > 30 or timeDelta_ka_received.total_seconds() > 30:
			return False
		else:
			return True


from typing import OrderedDict
from datetime import datetime
import pytz

from loguru import logger
class TriggerType:
	External: str = 'external'
	Time: str = 'time'

class TemplateTrigger:
	def __init__(self, triggerdata: OrderedDict) -> None:
		type = triggerdata['type']
		assert type in [TriggerType.External, TriggerType.Time]
		self.type = type
		self.value = triggerdata['value']
		assert self.value != '' and self.value != None
		
		try:
			timefrom = triggerdata['timefrom']
			self.fromTimeUTC = self.parseTimeWithTimezone(timefrom).astimezone(pytz.utc)
		except KeyError:
			self.fromTimeUTC = None
		except ValueError as valError:
			logger.error(f'Invalid time format for parameter "timefrom"!')

		try:
			timeto = triggerdata['timeto']
			self.toTimeUTC = self.parseTimeWithTimezone(timeto).astimezone(pytz.utc)
		except KeyError:
			self.toTimeUTC = None
		except ValueError as valError:
			logger.error(f'Invalid time format for parameter "timeto"!')

		try:
			excludeTimeFrom = triggerdata['excludetimefrom']
			self.excludeFromTimeUTC = self.parseTimeWithTimezone(excludeTimeFrom).astimezone(pytz.utc)
		except KeyError:
			self.excludeFromTimeUTC = None
		except ValueError as valError:
			logger.error(f'Invalid time format for parameter "excludetimefrom"!')

		try:
			excludeTimeTo = triggerdata['excludetimeto']
			self.excludeToTimeUTC = self.parseTimeWithTimezone(excludeTimeTo).astimezone(pytz.utc)
		except KeyError:
			self.excludeToTimeUTC = None
		except ValueError as valError:
			logger.error(f'Invalid time format for parameter "excludetimeto"!')
		
	def parseTimeWithTimezone(self, timeStr: str) -> datetime.time:
		""" 
		Parses a time string like 11:00 EST to a datetime.time object
		"""
		if timeStr == None:
			return None
		
		parts = timeStr.split(' ')
		try:
			now = datetime.now()
			naive_time = datetime.strptime(parts[0], '%H:%M').replace(year=now.year, month=now.month, day=now.day)
			if len(parts) > 1:
				tzstr = parts[1]
			else:
				tzstr = 'EST'
			tz = pytz.timezone(tzstr)
			localized_time = tz.localize(naive_time)
			return localized_time
		except ValueError and AttributeError as error:
			logger.error(f'Invalid time format: {timeStr} - Expecting HH:MM <Timezone>')
			return None

	def toDict(self):
		""" Returns a dictionary representation of the Trigger which is used for
		the config file.
		"""
		returnDict = {'type': self.type, 'value': self.value}
		eastern = pytz.timezone('US/Eastern')
		if self.fromTimeUTC != None:
			returnDict['timefrom'] = self.fromTimeUTC.astimezone(eastern).strftime('%H:%M %Z')
		if self.toTimeUTC != None:
			returnDict['timeto'] = self.toTimeUTC.astimezone(eastern).strftime('%H:%M %Z')
		if self.excludeFromTimeUTC != None:
			returnDict['excludetimefrom'] = self.excludeFromTimeUTC.astimezone(eastern).strftime('%H:%M %Z')
		if self.excludeToTimeUTC != None:
			returnDict['excludetimeto'] = self.excludeToTimeUTC.astimezone(eastern).strftime('%H:%M %Z')	
		return returnDict



from typing import List
from optrabot.models import Trade
from optrabot.tradetemplate.templatefactory import Template
from optrabot.broker.order import Execution, OptionRight, Order
from optrabot.stoplossadjuster import StopLossAdjuster
import copy

class ManagedTrade:
	"""
	ManagedTrade is representing a trade which is currently managed by the TradeManager.
	"""
	def __init__(self, trade: Trade, template: Template, entryOrder: Order, account: str = ''): 
		self.trade = trade
		self.entryOrder = entryOrder
		self.template = template
		self.account = account
		self.takeProfitOrder: Order = None
		self.stopLossOrder: Order = None
		self.status = 'NEW'
		self.realizedPNL = 0.0
		self.transactions = []
		self.expired = False
		self.entry_price = None					# Holds the entry price for the trade
		self.current_price: float = None		# Holds the current price of the trade
		self.stoploss_adjusters: List[StopLossAdjuster] = []
		self.long_legs_removed = False			# will be set to true for credit_trades if the long legs are no longer available

	def isActive(self) -> bool:
		"""
		Returns True if the trade is active
		"""
		return self.status == 'OPEN'
	
	def setup_stoploss_adjusters(self):
		""" 
		Copies the stop loss adjusters from the template to the managed trade and sets the
		base price for earch of the adjusters
		"""
		for adjuster in self.template.get_stoploss_adjusters():
			adjuster_copy = copy.copy(adjuster)
			adjuster_copy.setBasePrice(self.entry_price)
			self.stoploss_adjusters.append(adjuster_copy)


import array
from collections import OrderedDict
import asyncio
import datetime as dt
import json
from fastapi import FastAPI
from ib_async import *
import logging
from loguru import logger
import pytz
from sqlalchemy.orm import Session
from optrabot.broker.brokerconnector import BrokerConnector
from optrabot import schemas
from optrabot.broker.brokerfactory import BrokerFactory
from optrabot.marketdatatype import MarketDataType
from optrabot.optionhelper import OptionHelper
import optrabot.config as optrabotcfg
from optrabot.trademanager import TradeManager
from optrabot.tradetemplate.templatefactory import Template
import optrabot.symbolinfo as symbolInfo
from .tradinghubclient import TradinghubClient
import pkg_resources
from .database import *
from . import crud
from apscheduler.schedulers.asyncio import AsyncIOScheduler

def get_version() -> str:
	"""
	Returns the version of the package
	"""
	try:
		return pkg_resources.get_distribution('optrabot').version
	except pkg_resources.DistributionNotFound:
		return '0.13.9' # Set Version to 0.13.9 for the local development environment

class OptraBot():
	def __init__(self, app: FastAPI):
		self.app = app
		self._apiKey = None
		self.thc : TradinghubClient = None
		self._tradingEnabled = False
		self._marketDataType : MarketDataType = None
		self.Version = get_version()
		self._backgroundScheduler = AsyncIOScheduler()
		self._backgroundScheduler.start()
		#logging.getLogger('apscheduler').setLevel(logging.ERROR) # Prevents unnecessary logging from apscheduler
			
	def __setitem__(self, key, value):
		setattr(self, key, value)

	def __getitem__(self, key):
		return getattr(self, key)
	
	async def startup(self):
		logger.info('OptraBot {version}', version=self.Version)
		# Read Config
		conf = optrabotcfg.Config("config.yaml")
		optrabotcfg.appConfig = conf
		self['config'] = conf
		conf.logConfigurationData()
		conf.readTemplates()
		updateDatabase()
		self.thc = TradinghubClient(self)
		if self.thc._apiKey == None:
			return

		try:
			additional_data = {
				'instance_id': conf.getInstanceId(),
				'accounts': self._getConfiguredAccounts()
			}
			await self.thc.connect(additional_data)
		except Exception as excp:
			logger.error('Problem on Startup: {}', excp)
			logger.error('OptraBot halted!')
			return
		
		logger.info('Sucessfully connected to OptraBot Hub')
		await BrokerFactory().createBrokerConnectors()
		self.thc.start_polling(self._backgroundScheduler)
		TradeManager()
		self._backgroundScheduler.add_job(self._statusInfo, 'interval', minutes=5, id='statusInfo', misfire_grace_time=None)
		self._backgroundScheduler.add_job(self._new_day_start, 'cron', hour=0, minute=0, second=0, timezone=pytz.timezone('US/Eastern'), id='day_change', misfire_grace_time=None)

	async def shutdown(self):
		logger.info('Shutting down OptraBot')
		await self.thc.shutdown()
		TradeManager().shutdown()
		await BrokerFactory().shutdownBrokerConnectors()
		self._backgroundScheduler.shutdown()

	async def _new_day_start(self):
		"""
		Perform operations on start of a new day
		"""
		logger.debug('Performing Day Change operations')
		await BrokerFactory().new_day_start()

	def _statusInfo(self):
		siTradingEnabled = 'Yes' if self._tradingEnabled == True else 'No' 
		siPosition = 'Yes' if self.thc._position == True else 'No'
		siHubConnection = 'OK' if self.thc.isHubConnectionOK() == True else 'Problem!'

		managedTrades = TradeManager().getManagedTrades()
		activeTrades = 0
		for managedTrade in managedTrades:
			if managedTrade.isActive():
				activeTrades += 1

		logger.info(f'Status Info: Hub Connection: {siHubConnection} - Active Trades: {activeTrades}')

	def getMarketDataType(self) -> MarketDataType:
		""" Return the configured Market Data Type
		"""
		if self._marketDataType is None:
			config: Config = self['config']
			try:
				confMarketData = config.get('tws.marketdata')
			except KeyError as keyError:
				confMarketData = 'Delayed'
			self._marketDataType = MarketDataType()
			self._marketDataType.byString(confMarketData)
		return self._marketDataType
	
	async def _checkMarketData(self):
		""" Checks if the Market Data Subscription is as configured.
			It requests SPX Options Market Data and checks if the returned Market Data Type
			is Live Market data. If not, trading is prevented.
		"""
		self._tradingEnabled = False
		ib: IB = self['ib']
		if not ib.isConnected():
			return

		marketDataType = self.getMarketDataType()
		logger.debug("Requesting '{}' data from Interactive Brokers", marketDataType.toString())
		ib.reqMarketDataType(marketDataType.Value)

		symbolInformation = symbolInfo.symbol_infos['SPX']
		spx = Index('SPX', symbolInformation.exchange)
		await ib.qualifyContractsAsync(spx)
		for i in range(3):
			[ticker] = await ib.reqTickersAsync(spx)
			ibMarketDataType = MarketDataType(ticker.marketDataType)
			if ibMarketDataType.Value != marketDataType.Value:
				logger.error("IB returned '{}' data for SPX! Trading is deactivated!", ibMarketDataType.toString())
				return
			else:
				logger.info("Received '{}' market data for SPX as expected.", ibMarketDataType.toString())

			logger.debug("Ticker data: Last={} Close={} Market Price={}", ticker.last, ticker.close, ticker.marketPrice())	
			spxPrice = ticker.close
			if util.isNan(spxPrice):
				logger.debug("IB returned no SPX price but just NaN value for last price. Trading is deactivated!")
			else:
				break # no more loop required

		if util.isNan(spxPrice):
			logger.error("IB returned no SPX price, but just NaN value after 3 attempts! Trading is deactivated!")
			return

		chains = await ib.reqSecDefOptParamsAsync(spx.symbol, '', spx.secType, spx.conId)
		chain = next(c for c in chains if c.tradingClass == symbolInformation.trading_class and c.exchange == symbolInformation.exchange)
		if chain == None:
			logger.error("No Option Chain for SPXW and SMARE found! Not able to trade SPX options!")
			return

		current_date = dt.date.today()
		expiration = current_date.strftime('%Y%m%d')

		if int(chain.expirations[0]) > int(expiration):
			logger.warning('There are no SPX options expiring today!')
			expiration = chain.expirations[0]

		strikePrice = OptionHelper.roundToStrikePrice(spxPrice)
		logger.info("Requesting Short Put price of strike {}", strikePrice)
		shortPutContract = Option(spx.symbol, expiration, strikePrice, 'P', symbolInformation.exchange, tradingClass = symbolInformation.trading_class)
		await ib.qualifyContractsAsync(shortPutContract)
		if not OptionHelper.checkContractIsQualified(shortPutContract):
			return
		ticker = None
		[ticker] = await ib.reqTickersAsync(shortPutContract)
		ibMarketDataType = MarketDataType(ticker.marketDataType)
		if ibMarketDataType.Value != marketDataType.Value:
			logger.error("IB returned '{}' data for SPX Option! Trading is deactivated!", ibMarketDataType.toString())
			return
		else:
			logger.info("Received '{}' market data for SPX Option as expected.", ibMarketDataType.toString())
		
		optionPrice = ticker.close
		if util.isNan(optionPrice):
			logger.error("IB returned no price for the SPX option but just a NaN value. Trading is deactivated!")
			return

		logger.success("Market Data subscription checks passed successfully. Options Trading is enabled.")
		self._tradingEnabled = True

	def isTradingEnabled(self) -> bool:
		""" Returns true if trading is enabled after market data subscription checks have passed.
		"""
		return self._tradingEnabled

	def _getConfiguredAccounts(self) -> list:
		""" 
		Returns a list of configured accounts
		"""
		#conf: Config = self['config']
		conf: Config = optrabotcfg.appConfig
		configuredAccounts = None
		for item in conf.getTemplates():
			template : Template = item
			if configuredAccounts == None:
				configuredAccounts = [template.account]
			else:
				if not template.account in configuredAccounts:
					configuredAccounts.append(template.account)
		return configuredAccounts

	@logger.catch
	def handleTaskDone(self, task: asyncio.Task):
		if not task.cancelled():
			taskException = task.exception()
			if taskException != None:
				logger.error('Task Exception occured!')
				raise taskException


from datetime import datetime
from enum import Enum
from typing import Optional

class OrderAction(str, Enum):
    """
    Represents actions of an order or order le
    """

    BUY_TO_OPEN = "Buy to Open"
    BUY_TO_CLOSE = "Buy to Close"
    SELL_TO_OPEN = "Sell to Open"
    SELL_TO_CLOSE = "Sell to Close"
    BUY = "Buy"
    SELL = "Sell"

class OrderType(str, Enum):
	"""
	Represents order types
	"""
	LIMIT = "Limit"
	MARKET = "Market"
	STOP = "Stop"
	STOP_LIMIT = "Stop Limit"

class OrderStatus(str, Enum):
	"""
	Represents the status of an order
	"""
	OPEN = "Open"
	FILLED = "Filled"
	CANCELLED = "Cancelled"

class OptionRight(str, Enum):
	"""
	Represents the right of an option
	"""
	CALL = "Call"
	PUT = "Put"

class Leg():
	"""
	Represents a leg of an order.
	"""
	action: OrderAction
	quantity: Optional[int]
	symbol: str
	strike: float
	right: OptionRight
	expiration: datetime
	askPrice: float = None
	bidPrice: float = None
	midPrice: float = None
	brokerSpecific = {}
     
	def __init__(self, action: OrderAction, symbol: str, quantity: Optional[int] = 0, strike: float = None, right: OptionRight = None, expiration: datetime = None):
		self.action = action
		self.quantity = quantity
		self.symbol = symbol
		self.right = right
		self.strike = strike
		self.expiration = expiration

class PriceEffect(str, Enum):
    """
    This is an :class:`~enum.Enum` that shows the sign of a price effect.
    """
    CREDIT = "Credit"
    DEBIT = "Debit"
    NONE = "None"

class Order():
	"""
	Represents an order. A order can have multiple legs
	"""
	
	def __init__(self, symbol: str = '', action: OrderAction = None, type: OrderType = None, legs: list[Leg] = None, quantity: int = 1, price: float = None, price_effect: str = PriceEffect.NONE):
		self.symbol = symbol
		self.action = action
		self.legs: list[Leg] = legs
		self.type = type
		self.quantity = quantity
		self.price: float = price
		self.averageFillPrice: float = None
		self.ocaGroup = None
		self.status: OrderStatus = None
		self.orderReference: str = ''
		self.brokerSpecific = {}
		self.broker_order_id = ''
		self.price_effect = price_effect

class Execution():
	"""
	Represents a fill information of an order execution
	"""
	def __init__(self, id: str, action: OrderAction, sec_type: str, strike: float, amount: int, price: float, timestamp: datetime, expiration: datetime):
		self.id: str = id
		self.action: OrderAction =action
		self.sec_type: str = sec_type
		self.strike: float = strike
		self.timestamp: datetime = timestamp
		self.expiration: datetime = expiration
		self.amount: int = amount
		self.price: float = price